# -*- coding: utf-8 -*-
from .model_manager import ModelRegistry
from fosforml.utils import get_model_name, validate_datasets_for_sf
import snowflake,sys
import warnings
warnings.filterwarnings("ignore")

def register_model(
    model_obj,
    session,
    name,
    description,
    dataset_name=None,
    dataset_source=None,
    scoring_func=None,
    flavour='snowflake',
    model_type=None,
    snowflake_df=None,
    conda_dependencies=[],
    x_train=None,
    y_train=None,
    x_test=None,
    y_test=None,
    y_pred=None,
    y_prob=None,
    prediction_column=None,
    target_column=None,
    pretty_output=False,
    metadata=None,
    source=None,
    **kwargs
):
    """
    Register model to the Snowflake Model Registry

    Args:
        model_obj (object): The model object to be registered.
        session (snowflake.snowpark.session.Session): The Snowflake session object.
        scoring_func (function): The function to be used for scoring the model.
        name (str): The name of the model.
        description (str): The description of the model.
        snowflake_df (snowflake.snowpark.dataframe.DataFrame): The Snowflake DataFrame object. which should container prediction columns,feature columns and target column.
        flavour (str): The flavour of the model, e.g., keras, pytorch, tensorflow, etc.
        y_true (array, optional): The true labels of the model's predictions. Shape: [n_samples].
        y_pred (array, optional): The predicted labels of the model. Shape: [n_samples].
        prob (array-like, optional): The predicted probabilities of the model. Shape: (n_samples,).
        model_type (str, optional): The type of the model, e.g., classification, regression, etc.
        conda_dependencies (list, optional): The Conda dependencies required by the model.
        x_train (numpy array, optional): The training data of the model with feature columns.
        x_test (numpy array, optional): The test data of the model with feature columns.
        y_train (numpy array, optional): The training data of the model with target column.
        y_test (numpy array, optional): The test data of the model with target column.
        source (str, optional): The source of the Notebook,Experment, e.g., AutoML, etc.
        pretty_output (bool, optional): If True, returns a widget after registration; else, returns a dictionary.
        metadata (dict, optional): Metadata information about the version.

    Keyword Args:
        explicit_x_train (pd.DataFrame or np.ndarray, optional): Explicit x_train clean raw data.
            If provided, the following algorithms will automatically use it for their executions:
            - Know Your Data: explicit_x_train is used for extraction of knowledge. x_train is still used internally.
        explicit_x_test (pd.DataFrame or np.ndarray, optional): Explicit x_test clean raw data.
            If provided, the following algorithms will automatically use it for their executions:
            - Know Your Data: explicit_x_test is used for extraction of knowledge. x_test is still used internally.
        explicit_feature_names (list, optional): Feature names for the explicit provided data.
        source (str, optional): Value will be "automl" if the model is registered from AutoML; else, None.
        model_display (bool, optional): If True, display the model on the model list.

    Returns:
        The response from the model registry.

    Raises:
        ValueError: If the session object is invalid.

    """

    if flavour is None:
        return False,"Model flavour is required, please provide the value for 'flavour' ."
    
    is_invalid_dfs,validation_response = validate_datasets_for_sf(model_obj,name,model_type,flavour,snowflake_df,
                                                       x_train,y_train,x_test,
                                                       y_test,y_pred,y_prob,dataset_name,dataset_source,
                                                       prediction_column)
    if is_invalid_dfs is False:
        return validation_response

    status,model_name = get_model_name(
        name=name,
        source=source
    )
    if not status:
        return model_name

    model_display = kwargs.get("model_display")
    if model_display is None:
        model_display = True
    
    if not session or not isinstance(session,snowflake.snowpark.session.Session):
        return False,"Invalid session object."

    # register model
    model_registry = ModelRegistry(session=session)
    reg_status,response = model_registry.register_model(
        model=model_obj,
        score=None,
        model_name=model_name,
        dataset_name=dataset_name,
        dataset_source=dataset_source,
        conda_dependencies=conda_dependencies,
        description=description,
        model_flavour=flavour,
        model_type=model_type,
        sf_input_dataframe=snowflake_df,
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        y_pred=y_pred,
        prob=y_prob,
        prediction_column=prediction_column,
        target_column=target_column,
        python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
        source=source,
        metadata=metadata
    )

    return response



# -*- coding: utf-8 -*-
import numpy
import pytest
import requests
from PIL import Image
from sklearn import datasets, svm

from fosforml import scoring_func

def _env_setup(monkeypatch):
    monkeypatch.setenv(
        "base_docker_image_name",
        "python",
    )

 

@pytest.fixture(autouse=True)
def env_setup(monkeypatch):
    _env_setup(monkeypatch)

@pytest.fixture(autouse=True)
def model_dict():
    return {
        "created_by": "10662254",
        "created_on": "2019-08-06T08:40:59.273662+00:00",
        "description": "svm classifier using sklearn",
        "flavour": "sklearn",
        "id": "c6252ff7-8928-4d43-b433-02fb67f24a3d",
        "init_script": "",
        "last_modified_by": "10662254",
        "last_modified_on": None,
        "name": "Credit_Train",
        "project_id": "1",
        "versions": [
            {
                "created_by": "10662254",
                "created_on": "2019-08-06T08:58:27.519680+00:00",
                "description": None,
                "id": "63d4ce07-ee31-4c1b-9bb1-52fb0c1901c9",
                "last_modified_by": "10662254",
                "last_modified_on": None,
                "ml_model_id": "c6252ff7-8928-4d43-b433-02fb67f24a3d",
                "object_url": None,
            }
        ],
    }


@pytest.fixture(autouse=True)
def no_deployments_model_dict():
    return {
        "created_by": "10662254",
        "created_on": "2019-08-06T08:40:59.273662+00:00",
        "description": "svm classifier using sklearn",
        "flavour": "sklearn",
        "id": "c6252ff7-8928-4d43-b433-02fb67f24a3d",
        "init_script": "",
        "last_modified_by": "10662254",
        "last_modified_on": None,
        "name": "Credit_Train",
        "project_id": "1",
        "versions": [
            {
                "created_by": "10662254",
                "created_on": "2019-08-06T08:58:27.519680+00:00",
                "description": None,
                "deployments": [],
                "id": "63d4ce07-ee31-4c1b-9bb1-52fb0c1901c9",
                "last_modified_by": "10662254",
                "last_modified_on": None,
                "ml_model_id": "c6252ff7-8928-4d43-b433-02fb67f24a3d",
                "object_url": None,
            }
        ],
    }


@pytest.fixture(autouse=True)
def deployment_model_dict_deployed():
    return {
        "created_by": "mosaic.ai",
        "created_on": "2020-01-28T09:37:58+00:00",
        "deploymentstatus": True,
        "description": "svm classifier using sklearn",
        "flavour": "sklearn",
        "id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
        "last_modified_by": "mosaic.ai",
        "last_modified_on": "2020-01-28T09:37:58+00:00",
        "name": "TestingGiven",
        "project_id": "1",
        "type": "model",
        "versions": [
            {
                "created_by": "mosaic.ai",
                "created_on": "2020-01-28T09:39:53+00:00",
                "datasource_name": None,
                "deploy_info": None,
                "deployments": [
                    {
                        "created_by": "mosaic.ai",
                        "created_on": "2020-08-25T17:59:12+00:00",
                        "deployment_info": {
                            "configmap_name": "jwt-dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "cpu_utilization": "75",
                            "deployment_name": "dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "deployment_type": "Default",
                            "host_name": "ratanboddu.com/testingartinofucksgiven/45c7e7db-5674-4878-bd40-59c38a9521f2",
                            "hpa_name": "hpa-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "ingress_name": "ing-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "init_configmap_name": "script-dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "labels": {},
                            "service_name": "svc-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "strategy_deployment_id": "46f7a0b7-94ab-483e-aee1-9d39242441cd",
                            "strategy_version_id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                        },
                        "deployment_status": "DEPLOYED",
                        "id": "19430fc7-89e7-4e78-be2f-3fd5b01c81a6",
                        "last_modified_by": "mosaic.ai",
                        "last_modified_on": "2020-08-26T09:29:53+00:00",
                        "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
                        "name": "dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                        "resource": {
                            "cpu_limit": "500m",
                            "cpu_request": "50m",
                            "gpu": "",
                            "gpu_type": None,
                            "id": "b0d93c79-b4f0-46c0-8a54-e41abbaa5583",
                            "max_replicas": "3",
                            "min_replicas": "1",
                            "name": "Micro",
                            "ram_limit": "300Mi",
                            "ram_request": "50Mi",
                        },
                        "resource_id": "b0d93c79-b4f0-46c0-8a54-e41abbaa5583",
                        "version_id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                    }
                ],
                "description": None,
                "docker_image_url": "registry.lti-aiq.in:443/mosaic-ai-logistics/mosaic-ai-serving:1.0.0-52280",
                "id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                "init_script": '"pip install absl-py==0.9.0 appdirs==1.4.3 appnope==0.1.0 aspy.yaml==1.3.0 astor==0.8.1 attrs==19.3.0 backcall==0.1.0 black==19.10b0 bleach==3.1.0 blis==0.4.1 cachetools==4.0.0 certifi==2019.11.28 cffi==1.13.2 cfgv==2.0.1 chardet==3.0.4 Click==7.0 cloudpickle==1.2.1 coverage==5.0.2 cymem==2.0.3 decorator==4.4.1 defusedxml==0.6.0 entrypoints==0.3 gast==0.3.2 google-auth==1.10.0 google-pasta==0.1.8 grpcio==1.26.0 h5py==2.10.0 identify==1.4.9 idna==2.8 importanize==0.7.0 importlib-metadata==1.3.0 importlib-resources==1.0.2 ipykernel==5.1.3 ipython==7.11.1 ipython-genutils==0.2.0 ipywidgets==7.5.1 jedi==0.15.2 Jinja2==2.10.3 joblib==0.14.1 jsonschema==3.2.0 jupyter==1.0.0 jupyter-client==5.3.4 jupyter-console==6.0.0 jupyter-core==4.6.1 Keras==2.3.1 Keras-Applications==1.0.8 Keras-Preprocessing==1.1.0 kubernetes==9.0.0 Markdown==3.1.1 MarkupSafe==1.1.1 mistune==0.8.4 more-itertools==8.0.2 murmurhash==1.0.2 nbconvert==5.6.1 nbformat==5.0.3 nodeenv==1.3.4 notebook==6.0.2 numpy==1.17.0 oauthlib==3.1.0 packaging==20.0 pandas==0.25.1 pandocfilters==1.4.2 parso==0.5.2 pathlib2==2.3.5 pathspec==0.7.0 pexpect==4.7.0 pickleshare==0.7.5 Pillow==7.0.0 pip==19.3.1 plac==1.1.3 pluggy==0.13.1 pre-commit==1.21.0 preshed==3.0.2 prometheus-client==0.7.1 prompt-toolkit==2.0.10 protobuf==3.11.2 ptyprocess==0.6.0 py==1.8.1 py4j==0.10.4 pyasn1==0.4.8 pyasn1-modules==0.2.8 pycparser==2.19 Pygments==2.5.2 pyparsing==2.4.6 pyrsistent==0.15.7 pyspark==2.1.2 pytest==5.3.2 pytest-cov==2.8.1 python-dateutil==2.8.1 pytz==2019.3 PyYAML==5.1.2 pyzmq==18.1.1 qtconsole==4.6.0 regex==2020.1.8 requests==2.22.0 requests-oauthlib==1.3.0 rpy2==3.2.0 rsa==4.0 scikit-learn==0.21.3 scipy==1.4.1 Send2Trash==1.5.0 setuptools==44.0.0 simplegeneric==0.8.1 six==1.13.0 sklearn==0.0 spacy==2.2.2 srsly==1.0.1 tensorboard==1.14.0 tensorflow==1.14.0 tensorflow-estimator==1.14.0 termcolor==1.1.0 terminado==0.8.3 testpath==0.4.4 thinc==7.3.1 toml==0.10.0 torch==1.2.0 torchvision==0.4.0 tornado==6.0.3 tqdm==4.41.1 traitlets==4.3.3 typed-ast==1.4.0 tzlocal==2.0.0 urllib3==1.25.7 virtualenv==16.7.9 wasabi==0.6.0 wcwidth==0.1.8 webencodings==0.5.1 websocket-client==0.57.0 Werkzeug==0.16.0 wheel==0.33.6 widgetsnbextension==3.5.1 wrapt==1.11.2 zipp==0.6.0 \\nmkdir abc"',
                "input_type": "json",
                "last_modified_by": "mosaic.ai",
                "last_modified_on": "2020-01-28T09:39:54+00:00",
                "metadata_info": None,
                "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
                "model_class": None,
                "model_info": None,
                "object_url": "45c7e7db-5674-4878-bd40-59c38a9521f2/28617a40-64b7-4dfc-a0d1-35eef9f2c558/ml_model.tar.gz",
                "profiling": [],
                "schema": None,
                "status": "active",
                "target_names": None,
                "version_no": 2,
            },
            {
                "created_by": "mosaic.ai",
                "created_on": "2020-01-28T09:37:58+00:00",
                "datasource_name": None,
                "deploy_info": None,
                "deployments": [
                    {
                        "created_by": "mosaic.ai",
                        "created_on": "2020-08-26T09:29:51+00:00",
                        "deployment_info": {
                            "configmap_name": "jwt-dp-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "cpu_utilization": "75",
                            "deployment_name": "dp-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "deployment_type": "PreProd",
                            "host_name": "mosaic.dev.lti-mosaic.com/testingartinofucksgiven/45c7e7db-5674-4878-bd40-59c38a9521f2",
                            "hpa_name": "hpa-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "ingress_name": "ing-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "init_configmap_name": "script-dp-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "labels": {},
                            "service_name": "svc-4250de43-ea10-4c92-881a-d01907f24b5b",
                            "strategy_deployment_id": "19430fc7-89e7-4e78-be2f-3fd5b01c81a6",
                            "strategy_version_id": "4250de43-ea10-4c92-881a-d01907f24b5b",
                        },
                        "deployment_status": "DEPLOYING",
                        "id": "46f7a0b7-94ab-483e-aee1-9d39242441cd",
                        "last_modified_by": "mosaic.ai",
                        "last_modified_on": "2020-08-26T09:29:53+00:00",
                        "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
                        "name": None,
                        "resource": {
                            "cpu_limit": "1",
                            "cpu_request": "50m",
                            "gpu": "",
                            "gpu_type": None,
                            "id": "844991ca-5d3f-4ddd-b9bf-34c9b78125f0",
                            "max_replicas": "3",
                            "min_replicas": "1",
                            "name": "default",
                            "ram_limit": "800Mi",
                            "ram_request": "50Mi",
                        },
                        "resource_id": "844991ca-5d3f-4ddd-b9bf-34c9b78125f0",
                        "version_id": "4250de43-ea10-4c92-881a-d01907f24b5b",
                    }
                ],
                "id": "4250de43-ea10-4c92-881a-d01907f24b5b",
                "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
            },
        ],
    }


@pytest.fixture(autouse=True)
def deployment_model_dict_deployed_VA():
    return {
        "created_by": "mosaic.ai",
        "created_on": "2020-01-28T09:37:58+00:00",
        "deploymentstatus": True,
        "description": "svm classifier using sklearn",
        "flavour": "sklearn",
        "id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
        "last_modified_by": "mosaic.ai",
        "last_modified_on": "2020-01-28T09:37:58+00:00",
        "name": "TestingGiven",
        "project_id": "1",
        "type": "model",
        "versions": [
            {
                "created_by": "mosaic.ai",
                "created_on": "2020-01-28T09:39:53+00:00",
                "datasource_name": None,
                "deploy_info": None,
                "deployments": [
                    {
                        "created_by": "mosaic.ai",
                        "created_on": "2020-08-25T17:59:12+00:00",
                        "deployment_info": {
                            "configmap_name": "jwt-dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "cpu_utilization": "75",
                            "deployment_name": "dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "deployment_type": "Default",
                            "host_name": "ratanboddu.com/testingartinofucksgiven/45c7e7db-5674-4878-bd40-59c38a9521f2",
                            "hpa_name": "hpa-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "ingress_name": "ing-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "init_configmap_name": "script-dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "labels": {},
                            "service_name": "svc-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                            "strategy_deployment_id": "46f7a0b7-94ab-483e-aee1-9d39242441cd",
                            "strategy_version_id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                        },
                        "deployment_status": "DEPLOYED",
                        "id": "19430fc7-89e7-4e78-be2f-3fd5b01c81a6",
                        "last_modified_by": "mosaic.ai",
                        "last_modified_on": "2020-08-26T09:29:53+00:00",
                        "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
                        "name": "dp-28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                        "resource": {
                            "cpu_limit": "500m",
                            "cpu_request": "50m",
                            "gpu": "",
                            "gpu_type": None,
                            "id": "b0d93c79-b4f0-46c0-8a54-e41abbaa5583",
                            "max_replicas": "3",
                            "min_replicas": "1",
                            "name": "Micro",
                            "ram_limit": "300Mi",
                            "ram_request": "50Mi",
                        },
                        "resource_id": "b0d93c79-b4f0-46c0-8a54-e41abbaa5583",
                        "version_id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                    }
                ],
                "description": None,
                "docker_image_url": "registry.lti-aiq.in:443/mosaic-ai-logistics/mosaic-ai-serving:1.0.0-52280",
                "id": "28617a40-64b7-4dfc-a0d1-35eef9f2c558",
                "init_script": '"pip install absl-py==0.9.0 appdirs==1.4.3 appnope==0.1.0 aspy.yaml==1.3.0 astor==0.8.1 attrs==19.3.0 backcall==0.1.0 black==19.10b0 bleach==3.1.0 blis==0.4.1 cachetools==4.0.0 certifi==2019.11.28 cffi==1.13.2 cfgv==2.0.1 chardet==3.0.4 Click==7.0 cloudpickle==1.2.1 coverage==5.0.2 cymem==2.0.3 decorator==4.4.1 defusedxml==0.6.0 entrypoints==0.3 gast==0.3.2 google-auth==1.10.0 google-pasta==0.1.8 grpcio==1.26.0 h5py==2.10.0 identify==1.4.9 idna==2.8 importanize==0.7.0 importlib-metadata==1.3.0 importlib-resources==1.0.2 ipykernel==5.1.3 ipython==7.11.1 ipython-genutils==0.2.0 ipywidgets==7.5.1 jedi==0.15.2 Jinja2==2.10.3 joblib==0.14.1 jsonschema==3.2.0 jupyter==1.0.0 jupyter-client==5.3.4 jupyter-console==6.0.0 jupyter-core==4.6.1 Keras==2.3.1 Keras-Applications==1.0.8 Keras-Preprocessing==1.1.0 kubernetes==9.0.0 Markdown==3.1.1 MarkupSafe==1.1.1 mistune==0.8.4 more-itertools==8.0.2 murmurhash==1.0.2 nbconvert==5.6.1 nbformat==5.0.3 nodeenv==1.3.4 notebook==6.0.2 numpy==1.17.0 oauthlib==3.1.0 packaging==20.0 pandas==0.25.1 pandocfilters==1.4.2 parso==0.5.2 pathlib2==2.3.5 pathspec==0.7.0 pexpect==4.7.0 pickleshare==0.7.5 Pillow==7.0.0 pip==19.3.1 plac==1.1.3 pluggy==0.13.1 pre-commit==1.21.0 preshed==3.0.2 prometheus-client==0.7.1 prompt-toolkit==2.0.10 protobuf==3.11.2 ptyprocess==0.6.0 py==1.8.1 py4j==0.10.4 pyasn1==0.4.8 pyasn1-modules==0.2.8 pycparser==2.19 Pygments==2.5.2 pyparsing==2.4.6 pyrsistent==0.15.7 pyspark==2.1.2 pytest==5.3.2 pytest-cov==2.8.1 python-dateutil==2.8.1 pytz==2019.3 PyYAML==5.1.2 pyzmq==18.1.1 qtconsole==4.6.0 regex==2020.1.8 requests==2.22.0 requests-oauthlib==1.3.0 rpy2==3.2.0 rsa==4.0 scikit-learn==0.21.3 scipy==1.4.1 Send2Trash==1.5.0 setuptools==44.0.0 simplegeneric==0.8.1 six==1.13.0 sklearn==0.0 spacy==2.2.2 srsly==1.0.1 tensorboard==1.14.0 tensorflow==1.14.0 tensorflow-estimator==1.14.0 termcolor==1.1.0 terminado==0.8.3 testpath==0.4.4 thinc==7.3.1 toml==0.10.0 torch==1.2.0 torchvision==0.4.0 tornado==6.0.3 tqdm==4.41.1 traitlets==4.3.3 typed-ast==1.4.0 tzlocal==2.0.0 urllib3==1.25.7 virtualenv==16.7.9 wasabi==0.6.0 wcwidth==0.1.8 webencodings==0.5.1 websocket-client==0.57.0 Werkzeug==0.16.0 wheel==0.33.6 widgetsnbextension==3.5.1 wrapt==1.11.2 zipp==0.6.0 \\nmkdir abc"',
                "input_type": "json",
                "last_modified_by": "mosaic.ai",
                "last_modified_on": "2020-01-28T09:39:54+00:00",
                "metadata_info": None,
                "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
                "model_class": None,
                "model_info": None,
                "object_url": "45c7e7db-5674-4878-bd40-59c38a9521f2/28617a40-64b7-4dfc-a0d1-35eef9f2c558/ml_model.tar.gz",
                "profiling": [],
                "schema": None,
                "status": "active",
                "target_names": None,
                "version_no": 2,
            },
            {
                "created_by": "mosaic.ai",
                "created_on": "2020-01-28T09:37:58+00:00",
                "datasource_name": None,
                "deploy_info": None,
                "deployments": [],
                "id": "4250de43-ea10-4c92-881a-d01907f24b5b",
                "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
            },
        ],
    }


@pytest.fixture(autouse=True)
def deployment_model_dict_deployed_None():
    return {
        "created_by": "mosaic.ai",
        "created_on": "2020-01-28T09:37:58+00:00",
        "deploymentstatus": True,
        "description": "svm classifier using sklearn",
        "flavour": "sklearn",
        "id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
        "last_modified_by": "mosaic.ai",
        "last_modified_on": "2020-01-28T09:37:58+00:00",
        "name": "TestingGiven",
        "project_id": "1",
        "type": "model",
        "versions": [
            {
                "created_by": "mosaic.ai",
                "created_on": "2020-01-28T09:37:58+00:00",
                "datasource_name": None,
                "deploy_info": None,
                "deployments": [],
                "id": "4250de43-ea10-4c92-881a-d01907f24b5b",
                "ml_model_id": "45c7e7db-5674-4878-bd40-59c38a9521f2",
            }
        ],
    }


@pytest.fixture(autouse=True)
def version_dict():
    return {
        "created_by": "10662254",
        "created_on": "2019-08-06T08:58:27.519680+00:00",
        "description": None,
        "id": "63d4ce07-ee31-4c1b-9bb1-52fb0c1901c9",
        "last_modified_by": "10662254",
        "last_modified_on": None,
        "ml_model_id": "c6252ff7-8928-4d43-b433-02fb67f24a3d",
        "object_url": None,
    }


@pytest.fixture(autouse=True)
def score_func():
    @scoring_func
    def score_func1(model, request):
        payload = request.json["payload"]
        data_list = payload
        data_array = numpy.asarray(data_list)
        prediction = model.predict(data_array)
        return prediction.tolist()

    return score_func1


@pytest.fixture(autouse=True)
def model_obj():
    digits = datasets.load_digits()

    clf = svm.SVC(gamma=0.001, C=100.0)
    clf.fit(digits.data[:-1], digits.target[:-1])
    return clf


@pytest.fixture(autouse=True)
def image_obj():
    image = Image.new(mode="RGB", size=(200, 200))
    image.save("/tmp/mosaic.jpg")
    image_object = Image.open("/tmp/mosaic.jpg")
    return image_object


@pytest.fixture(autouse=True)
def req_payload():
    req = requests.Request()
    req.json = {
        "payload": [
            [
                0.0,
                0.0,
                10.0,
                14.0,
                8.0,
                1.0,
                0.0,
                0.0,
                0.0,
                2.0,
                16.0,
                14.0,
                6.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                15.0,
                15.0,
                8.0,
                15.0,
                0.0,
                0.0,
                0.0,
                0.0,
                5.0,
                16.0,
                16.0,
                10.0,
                0.0,
                0.0,
                0.0,
                0.0,
                12.0,
                15.0,
                15.0,
                12.0,
                0.0,
                0.0,
                0.0,
                4.0,
                16.0,
                6.0,
                4.0,
                16.0,
                6.0,
                0.0,
                0.0,
                8.0,
                16.0,
                10.0,
                8.0,
                16.0,
                8.0,
                0.0,
                0.0,
                1.0,
                8.0,
                12.0,
                14.0,
                12.0,
                1.0,
                0.0,
            ]
        ]
    }
    return req.json


@pytest.fixture(autouse=True)
def model_ensemble_dict():
    return [
        {
            "model_id": "bbeb53ea-fa18-496b-bb4a-a6bd44b6508b",
            "version_id": "3878ea9e-8617-4b69-831b-2d36238e79d3",
            "name": "breastCancer_ensemble",
            "flavour": "sklearn",
        },
        {
            "model_id": "8a7c7a02-6a58-4e4d-ae93-db9c4c998fe7",
            "version_id": "7a909086-7ec8-4a8c-8efe-76e675852458",
            "name": "sklearn_ensemble",
            "flavour": "sklearn",
        },
    ]




import pandas as pd

class ModelObject:
    def __init__(self, model,model_type,snowflake_model=True):
        self.model_type = model_type
        self.model_obj = model
        self.model = model
        self.snowflake_model = snowflake_model
    
    def get_model_artifacts(self, session, snowflake_df, x_test, y_test, x_train, y_train, y_pred, y_prob):
        model_artifacts = {}
        new_model_object = self.get_model_object()
        model_artifacts['model_obj'] = new_model_object
        model_artifacts['hyper_parameters'] = self.get_hyper_parameters(new_model_object)
        model_artifacts['final_df'] = self.get_final_df(session, snowflake_df, x_test, y_test, x_train, y_train, y_pred, y_prob)
        return model_artifacts
    
    def get_hyper_parameters(self,new_model_object):
        return new_model_object.get_params()

    def get_model_object(self):
        if str(type(self.model_obj)).find("snowflake") > 1 :
            return self.model_obj.to_sklearn()

        if str(type(self.model_obj)).find("xgboost") >= 1 :
            return self.model_obj.to_xgboost()

        elif str(type(self.model_obj)).find("pipeline") >= 1 :
            temp_model = self.model_obj.steps[0][1]
            if str(type(temp_model)).find("xgboost") >= 1:
                return temp_model.to_xgboost()
            else:
                return temp_model.to_sklearn()
        else:
            return self.model_obj


    def get_final_df(self, session, snowflake_df, x_test, y_test, x_train, y_train, y_pred, y_prob):
        if self.snowflake_model:
            return snowflake_df
        else:
            no_rows = x_test.shape[0] ; final_pandas_dataframe = None
            if y_prob is None:
                final_pandas_dataframe = pd.concat([x_test.reset_index(drop=True).iloc[:no_rows,:],
                                                    y_test.reset_index(drop=True).squeeze(),
                                                    y_pred.reset_index(drop=True).squeeze()
                                                    ],axis=1)
            if isinstance(y_prob, pd.DataFrame):
                final_pandas_dataframe = pd.concat([x_test.reset_index(drop=True).iloc[:no_rows,:],
                                                    y_test.reset_index(drop=True).squeeze(),
                                                    y_pred.reset_index(drop=True).squeeze(),
                                                    y_prob.reset_index(drop=True).squeeze()
                                                    ],axis=1)

            final_pandas_dataframe_columns = final_pandas_dataframe.columns.to_list()
            return session.create_dataframe(final_pandas_dataframe,schema=final_pandas_dataframe_columns)

    





# -*- coding: utf-8 -*-
import os
connection_manager_url = os.getenv("CONNECTION_MANAGER_BASE_URL", "http://fdc-project-manager:80/project-manager")


class ModelConstants:
    snowflake_model_flavour = "snowflake"
    sklearn_model_flavour = "sklearn"

class UserConfigs:
    repo_key = "repo_details"
    connection_details = "connection_details"
    run_id = "RUN_ID"
    project_id = "PROJECT_ID"

class MosaicAI:
    server = os.getenv("MOSAIC_AI_SERVER", "http://localhost:5000/registry/api")
    kyd_setting = "/v1/ml-model/{}/version/{}/kyd/settings"

class NotebooksAPI:
    notebooks_api_server_url = os.getenv("NOTEBOOKS_API_SERVER", "http://notebooks-api.insight.svc.cluster.local:5000/notebooks/api")
    git_repo = os.getenv("GIT_REPO", "/v1/git-repo")


class RPYPackage:
    r_url = os.getenv("R_PACKAGE_REPO", "http://cran.us.r-project.org")
    py_url = os.getenv("PYPI_PACKAGE_REPO", "https://pypi.org/simple")
    conda_url = os.getenv("CONDA_PACKAGE_REPO", "https://repo.anaconda.com/pkgs/main/")
    override_channel_value = (
        "override-channels"
        if os.getenv("ARTIFACTORY") and os.getenv("ARTIFACTORY").lower() == "true"
        else ""
    )


class MLModelV1:
    c = "/v1/ml-model/register"
    u = "/v1/ml-model/add-version"
    lc = "/v1/ml-model"
    rud = "/v1/ml-model/{ml_model_id}"
    rud_model_name = "/v1/ml-model/{ml_model_id}?ml_model_name={ml_model_name}"
    rud_model_name1 = (
        "/v1/ml-model/{ml_model_id}?ml_model_name={ml_model_name}&type={model_type}"
    )


class MLModelResource:
    r = "/v1/ml-model/resource"


class MLModelVersionV1:
    lc = "/v1/ml-model/{ml_model_id}/version"
    rud = "/v1/ml-model/{ml_model_id}/version/{version_id}"
    upload = "/v1/ml-model/{ml_model_id}/version/{version_id}/upload"
    download = "/v1/ml-model/{ml_model_id}/version/{version_id}/download"
    build_time_metrics = "/v1/ml-model/metrics/{version_id}/{tag}"
    delete = "/v1/ml-model/{ml_model_id}/version/{version_id}"


class MLModelMetricsV1:
    lc = "/v1/ml-model/metrics"
    rud = "/v1/ml-model/metrics/{version_id}/{tag}"


class MLModelArtifactsV1:
    add_artifacts_url = "/v1/ml-model/add-artifacts"
    download_artifacts_url = "/v1/model/request-logger/download"


class MLModelDeployV1:
    c = "/v1/ml-model/{ml_model_id}/deploy"
    ud = "/v1/ml-model/{ml_model_id}/deploy/{deployment_id}"


class Client:
    config_file = "~/.mosaic.ai"


class Headers:
    authorization = "Authorization"
    x_project_id = "X-Project-Id"


class Artifacts:
    model = "model.tar.gz"


class MLModelFlavours:
    keras = "keras"
    sklearn = "sklearn"
    pytorch = "pytorch"
    tensorflow = "tensorflow"
    pyspark = "pyspark"
    spacy = "spacy"
    r = "r"
    ensemble = "ensemble"
    xgboost = "xgboost"


class MLModelProfiling:
    l = "/v1/profiling/datasource/{datasource_name}/column-name/{column_name}/version/{version_id}/ml-model/{ml_model_id}"
    c = "/v1/profiling"


class MLModelType:
    classification = "classification"
    regression = "regression"


class MosaicPipPackages:
    client = "mosaic-ai-client"
    automl = "Mosaic-Auto-Ml"
    common_utils = "mosaic-common-utils"
    connector = "mosaic-connector-python"
    visual_client = "mosaic-visual-client"


class CRANPackageList:
    pre_installed_packages = [
        "base",
        "boot",
        "class",
        "cluster",
        "codetools",
        "compiler",
        "datasets",
        "evaluate",
        "foreign",
        "graphics",
        "grDevices",
        "grid",
        "IRkernel",
        "KernSmooth",
        "lattice",
        "MASS",
        "Matrix",
        "methods",
        "mgcv",
        "mosaicrml",
        "nlme",
        "nnet",
        "parallel",
        "pbdZMQ",
        "compareDF",
        "rpart",
        "spatial",
        "splines",
        "stats",
        "stats4",
        "survival",
        "tcltk",
        "tools",
        "utils",
    ]


class MLModelVersionMetadataInfo:
    u = "/v1/ml-model/{ml_model_id}/version/{version_id}/metadata"
    l = "/v1/ml-model/model-meta/{ml_model_id}/version/{version_id}"


class MLModelVersionListing:
    l = "/v1/ml-model/version/list"


class DeployFeasibility:
    default = "default_deploy"
    apply_strategy = "strategy_deploy"
    no_deploy = "no_deploy"


class MLModelVersionFeedback:
    r = "/v1/ml-model/metrics/version/{version_id}/model-feedback"


class MLKYDDataStoreV1:
    kyd = "/v1/ml-model/{ml_model_id}/version/{version_id}/kyd"


class KYDValidations:
    IS_DATA_SUPPORTED = "Supported Data Format"
    UNSUPPORTED_DATA_ERROR = "Know Your Data Does Not Support {} Data Formats"

    IS_SHAPE_CORRECT = "Number of Records Mismatch Validation"

    VALIDATION_FAILED = "KYD Cannot Be Executed Due To Validation Errors."


class KYDExecutorLabels(KYDValidations):
    SUPPORTED_FORMATS = ["NUMPY", "PANDAS"]
    PUBLISHED = "Know Your Data for /{}/{}/ Published"
    PUBLISHED_FAIL = "Unable to publish Know Your Data for /{}/{}/"
    CHUNK_SIZE = 200


class SparkDistributed:
    executor_pod_image = os.getenv("executor_pod_image")
    executor_request_cpu = os.getenv("executor_request_cpu")
    executor_request_memory = os.getenv("executor_request_memory")
    executor_limit_cpu = os.getenv("executor_limit_cpu")
    executor_limit_memory = os.getenv("executor_limit_memory")
    number_of_executors = os.getenv("number_of_executors")
    executor_service_name = os.getenv("pod_name")
    pvc_name = os.getenv("pvc_name")
    project_id = os.getenv("PROJECT_ID")
    user_id = os.getenv("userId")
    minio_data_bucket = os.getenv("MINIO_DATA_BUCKET")
    namespace = os.getenv("NAMESPACE")
    is_job_run = os.getenv("is_job_run")
    PYTHONPATH = os.getenv("PYTHONPATH")



from snowflake.ml.dataset import Dataset
from fosforml.model_manager import snowflakesession as SnowflakeSession
import pandas as pd
import snowflake,json

class DatasetManager:
    def __init__(self,
                 model_name,
                 version_name,
                 session
                ):
        self.model_name = model_name
        self.version_name = version_name
        self.session_instance = SnowflakeSession()
        self.datasets_obj = self.get_or_create_dataset_object(model_name,version_name,session)
    
    def get_or_create_dataset_object(self,model_name,version_name,session):
        datasets_obj = Dataset(session=session,
                                database=self.session_instance.connection_params['database'],
                                schema =self.session_instance.connection_params['schema'],
                                name=f"{model_name}_{version_name}"
                                )
        try:
            datasets_obj.create(session=session,name=f"{model_name}_{version_name}")
        except Exception as e:
            datasets_obj.load(session=session,name=f"{model_name}_{version_name}")
        
        return datasets_obj
    
    def upload_datasets(self,session,datasets: dict):
        try:
            for dataset_name, dataset in datasets.items():
                if dataset is None:
                    continue
                snowpark_dataset = self.get_snowpark_dataset(session,dataset)
                purpose = self.get_dataset_purpose(dataset_name)
                self.datasets_obj.create_version(version=dataset_name,
                                                 input_dataframe=snowpark_dataset,
                                                 comment=json.dumps(
                                                            {
                                                                "purpose": purpose,
                                                                "dataset_type": "Table"
                                                            }
                                                        ))
                
            return True,f"Successfully uploaded {self.model_name} datasets."
        except Exception as e:
            print(e)
            raise Exception(e)
        
    def remove_datasets(self):
        try:
            self.datasets_obj.delete()
            return True,f"Successfully removed {self.model_name} datasets."
        except Exception as e:
            raise Exception(e)
    
    @staticmethod
    def get_dataset_purpose(dataset_name):
        if "x_train" in dataset_name.lower():
            return "Training"
        elif "y_train" in dataset_name.lower():
            return "Training"
        elif "x_test" in dataset_name.lower():
            return "Inference"
        elif "y_test" in dataset_name.lower():
            return "Inference"
        elif "y_pred" in dataset_name.lower():
            return "Validation"
        elif "prob" in dataset_name.lower():
            return "Validation"           
        
    def get_snowpark_dataset(self,session,dataset):
        if isinstance(dataset,snowflake.snowpark.dataframe.DataFrame):
            return dataset
        elif isinstance(dataset,pd.DataFrame):
            if dataset.empty:
                raise Exception("Empty dataframe provided. Please provide a non-empty dataframe.")
            return session.create_dataframe(dataset)
        elif isinstance(dataset,pd.Series):
            if dataset.empty:
                raise Exception("Empty series provided. Please provide a non-empty series.")
            return session.create_dataframe(dataset.to_list())
        else:
            raise Exception("Invalid dataset type. Please provide a pandas DataFrame or Series object.")    

class Metadata:
    def __init__(self, model_registry):
        self.model_registry = model_registry

    def update_model_registry(self,
                              model_name,
                              model_description,
                              model_tags,
                              session
                              ):
        try:
            model = self.model_registry.get_model(model_name=model_name)
            model.description = model_description
            self.set_model_tags(session,
                                model,
                                model_name,
                                tags=model_tags)
            return True, f"Successfully updated model metadata for {model_name}."
        except Exception as e:
            return False, f"Failed to update model metadata for {model_name}. {str(e)}"

    def set_model_tags(self,
                       session,
                       model,
                       model_name,
                       tags={}):
        try:
            for tag_name,tag_value in tags.items():
                session.sql(f"create tag if not exists {tag_name}").collect()
                model.set_tag(
                            tag_name = tag_name,
                            tag_value = tag_value
                            )
        except Exception as e:
            print(f"Failed to set tags for model {model_name}.")
            print(e)
            # raise Exception(f"Failed to set tags for model {model_name}")
        

    




from snowflake.ml.modeling.metrics import (confusion_matrix,
                                        accuracy_score,
                                        f1_score, recall_score,
                                        precision_score,
                                        log_loss ,
                                        roc_auc_score ,
                                        roc_curve)

from snowflake.snowpark.functions import sqrt ,log, abs
import json

def generate_metrics(model, combined_df,model_type,true_cn,pred_cn,pred_proba):
    """
    Method to generate metrics
    """
    cls = None
    if model_type.lower() == "classification":
        pass
        cls = Classification(model, combined_df,model_type,true_cn,pred_cn,pred_proba)
    elif model_type.lower() == "regression":
        cls = Regression(model, combined_df,model_type,true_cn,pred_cn,pred_proba)
    return cls.get_metrics()
        

def try_or(fn):
    try:
        out = fn()
        return out
    except Exception as e:
        print(e)
        return None

def update_progress(progress):
    bar_length = 70
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
    if progress < 0:
        progress = 0
    if progress >= 1:
        progress = 1
    block = int(round(bar_length * progress))
    print("Calculating build time metrics\n")
    text = "Progress: {0} {1:.1f}%".format(
        "█" * block + " " * (bar_length - block), progress * 100
    )
    print(text)



class Regression:

    def __init__(self, model, sf_df,true_cn,pred_cn,pred_proba_cn,sf_input_cols):
        """
        Constructor to initialize the model object
        """
        self.model = model
        self.sf_df = sf_df 
        self.true_cn = true_cn
        self.pred_cn = pred_cn 
        self.pred_proba_cn = pred_proba_cn
        self.progress_base_value = 4
        self.progress_counter = 1 
        self.sf_input_cols = sf_input_cols

    def get_metrics(self):
        """
        Method to get metrics 
        
        """
        all_summary = [
            self.cal_scatter_plot(),
            self.detailed_matrix(),
            self.decision_tree(self.model,self.sf_input_cols),
            self.cal_feature_importance(),
        ]

        return all_summary

    def cal_feature_importance(self):
        """
        Method to calculate variable importance
        """
        try:
            imp_list,feature_list,sklearn_model = [], None ,self.model

            if hasattr(self.model,"feature_importances_"):
                feature_list = self.model.feature_importances_
            elif hasattr(self.model,"coef_"):
                feature_list = self.model.coef_
            else:
                pass
                
            for index , value in zip(sklearn_model.feature_names_in_,feature_list.tolist()):
                imp_list.append({"column_name": index, "importance": value})
                
        except Exception:
            imp_list = None
        feature_importance_dict = {
            "tag": "feature_importance",
            "model_metric_value": imp_list,
        }
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return feature_importance_dict

    def cal_scatter_plot(self):
        """
            Method is used to calculate the scatter plot
        """
        try:
            values_1  = self.sf_df.group_by([self.true_cn[0],self.pred_cn[0]]).agg().limit(1000).collect()
            list1 = []
            for i,j in values_1:
                list1.append([i,j])
            scatt_dict = {"tag": "scatter_plot", "model_metric_value": list1}
            update_progress(self.progress_counter / self.progress_base_value)
            self.progress_counter += 1
            return scatt_dict
        except Exception as ex:
            print("Error while generating scatt plot")
    
    def detailed_matrix(self):
        
        """
        Method to calculate detailed matrix values of regression model
        """
        detailed_matrix_dict = {}
        import math
        try :
            from snowflake.ml.modeling.metrics import explained_variance_score, mean_absolute_percentage_error , mean_squared_error,mean_absolute_error,r2_score
            detailed_matrix_dict[explained_variance_score.__name__] = try_or(lambda : explained_variance_score(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            detailed_matrix_dict[mean_absolute_percentage_error.__name__] = try_or(lambda : mean_absolute_percentage_error(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            detailed_matrix_dict[mean_squared_error.__name__] = try_or(lambda : mean_squared_error(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            detailed_matrix_dict[mean_absolute_error.__name__] = try_or(lambda : mean_absolute_error(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            detailed_matrix_dict[r2_score.__name__] = try_or(lambda : r2_score(df=self.sf_df,y_true_col_name=self.true_cn[0], y_pred_col_name=self.pred_cn[0]))
            detailed_matrix_dict["Root Mean Square Error"] =  math.sqrt(detailed_matrix_dict[mean_squared_error.__name__])
            detailed_dict = {"tag": "detailed_matrix", "model_metric_value": detailed_matrix_dict}
            update_progress(self.progress_counter / self.progress_base_value)
            self.progress_counter += 1
        except Exception as ex:
            print("Error in while calculating detailed_matrix ")

        return detailed_dict

    def decision_tree(self,decision_tree, feature_names=None):
        decision_tree_ouput = {}
        try:
            if hasattr(decision_tree, "tree_"):
                from sklearn.tree import _tree
                
                tree_ = decision_tree.tree_
                feature_name = [
                    feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
                    for i in tree_.feature
                ]
                def recurse(node):
                    if tree_.feature[node] != _tree.TREE_UNDEFINED:
                        name = feature_name[node]
                        threshold = tree_.threshold[node]
                        left = recurse(tree_.children_left[node])
                        right = recurse(tree_.children_right[node])
                        return {"name": name, "threshold": threshold, "left": left, "right": right}
                    else:
                        return {"value": tree_.value[node].tolist()}
                
                decision_tree_ouput = {"tag": "decision_tree", "model_metric_value": recurse(0)}
            else:
                decision_tree_ouput = {"tag": "decision_tree", "model_metric_value": None}

        except Exception as msg:
            print("Error in while calculating decision_tree ")
        
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1

        return decision_tree_ouput

    

class Classification:

    def __init__(self, model_obj, sf_df,true_cn,pred_cn,pred_proba_cn,sf_input_cols):
        """
        Constructor to initialize the model object
        """
        self.sf_df = sf_df 
        self.true_cn = true_cn
        self.pred_cn = pred_cn 
        self.pred_proba_cn = pred_proba_cn
        self.progress_base_value = 5
        self.progress_counter = 1
        self.model = model_obj
        self.sf_input_cols= sf_input_cols
        self.no_classes=len(self.model.classes_)

    def get_metrics(self):
        """
        Method to get metrics  
        """ 

        all_summary = [
                self.confusion_metrics(),
                self.decision_tree(self.model,self.sf_input_cols),
                self.detailed_matrix(),
                self.cal_roc_auc(),
                self.cal_feature_importance()
        ]
        return all_summary

    def cal_feature_importance(self):
        """
        Method to calculate variable importance

        Returns:
            String
        """
        try:
            imp_list = []
            # sklearn_model = self.model.to_sklearn()
            for index , value in zip(self.model.feature_names_in_,self.model.feature_importances_.tolist()):
                imp_list.append({"column_name": index, "importance": value})
        except Exception as msg:
            print("Error in while calculating feature_importance ")
            imp_list = None
        feature_importance_dict = {
            "tag": "feature_importance",
            "model_metric_value": imp_list,
        }

        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return feature_importance_dict

    def confusion_metrics(self):
        """
        Method to return confusion matrix string

        Returns:
            String
        """
        try:
            dlist = []
            conf_matrix = confusion_matrix(df=self.sf_df, y_true_col_name=self.true_cn[0], y_pred_col_name=self.pred_cn[0])
            conf_list = conf_matrix.tolist()
            for idx, item in enumerate(conf_list):
                for index, val in enumerate(item):
                    d = {
                        "column_1_counter": idx,
                        "column_2_counter": index,
                        "prediction": val,
                        "column_1": idx,
                        "column_2": index,
                    }
                    dlist.append(d.copy())
            
        except Exception as ex:
            print("Error in while calculating confusion_matrix ",ex)
            dlist = None

        conf_dict = {"tag": "confusion_matrix", "model_metric_value": dlist}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return conf_dict

    def detailed_matrix(self):
        """
        Method to calculate detailed matrix parameters

        Returns:
            String
        """
        detailed_matrix_dict = {}
        average='weighted'
        # if hasattr(self.model,'n_classes_'):
        average='weighted' if self.no_classes > 2 else 'binary'

        detailed_dict = {}
        
        try :
            detailed_matrix_dict[accuracy_score.__name__] = try_or(lambda : accuracy_score(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            detailed_matrix_dict[precision_score.__name__] = try_or(lambda : precision_score(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn,average=average))
            detailed_matrix_dict[recall_score.__name__] = try_or(lambda : recall_score(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn,average=average))
            detailed_matrix_dict[f1_score.__name__] = try_or(lambda : f1_score(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn,average=average))
            detailed_matrix_dict[log_loss.__name__] = try_or(lambda : log_loss(df=self.sf_df,y_true_col_names=self.true_cn, y_pred_col_names=self.pred_cn))
            # detailed_matrix_dict[roc_auc_score.__name__]  =  try_or(lambda: roc_auc_score(df=self.sf_df,y_true_col_names=self.true_cn, y_score_col_names=self.pred_cn))
            detailed_dict = {"tag": "detailed_matrix", "model_metric_value": detailed_matrix_dict}
            update_progress(self.progress_counter / self.progress_base_value)
            self.progress_counter += 1
        except Exception as ex:
            print("Error in while calculating detailed_matrix ",ex)

        return detailed_dict

    def cal_roc_auc(self):
        """
        Method to calculate roc auc curve parameters

        Returns:
            String
        """
        try:
            roc_auc = None ; fpr = None ; tpr = None
            if self.no_classes > 2 and len(self.pred_proba_cn) == self.no_classes:
                roc_auc = roc_auc_score(df=self.sf_df, y_true_col_names=self.true_cn, y_score_col_names=self.pred_proba_cn, multi_class='ovr',average=None).tolist()
                fpr = None
                tpr = None
            elif self.no_classes == 2 :
                roc_auc = roc_auc_score(df=self.sf_df, y_true_col_names=self.true_cn, y_score_col_names=self.pred_cn ,average="macro").tolist()
                fpr, tpr, thresholds = roc_curve(df=self.sf_df,y_true_col_name=self.true_cn,y_score_col_name=self.pred_cn)
                fpr = fpr.tolist()
                tpr = tpr.tolist()
            else:
                roc_auc = None
                fpr = None
                tpr = None 
            
            roc_auc_value = {"fpr": fpr, "tpr": tpr, "data": roc_auc}
            
        except Exception as ex:
            print("Error in while calculating roc_auc ",ex)
            roc_auc_value = None

        roc_auc_dict = {"tag": "roc_auc", "model_metric_value": roc_auc_value}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return roc_auc_dict



    def decision_tree(self,decision_tree, feature_names=None):
        decision_tree_ouput = {}
        try:
            if hasattr(decision_tree, "tree_"):
                from sklearn.tree import _tree
                
                tree_ = decision_tree.tree_
                feature_name = [
                    feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
                    for i in tree_.feature
                ]
                def recurse(node):
                    if tree_.feature[node] != _tree.TREE_UNDEFINED:
                        name = feature_name[node]
                        threshold = tree_.threshold[node]
                        left = recurse(tree_.children_left[node])
                        right = recurse(tree_.children_right[node])
                        return {"name": name, "threshold": threshold, "left": left, "right": right}
                    else:
                        return {"value": tree_.value[node].tolist()}
                
                decision_tree_ouput = {"tag": "decision_tree", "model_metric_value": recurse(0)}
            else:
                decision_tree_ouput = {"tag": "decision_tree", "model_metric_value": None}

        except Exception as msg:
            print("Error in while calculating decision_tree ")
        
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1

        return decision_tree_ouput


# -*- coding: utf-8 -*-
import base64
import datetime
import json

import numpy as np
import pandas as pd
import pytz

from fosforml.schema.numpy_schema_generator import NumpySchemaGenerator
from fosforml.schema.pandas_schema_generator import PandasSchemaGenerator


def generate_service_schema(inputs, outputs):

    input_schema = _generate_obj_schema(inputs)
    output_schema = _generate_obj_schema(outputs)
    return {"input": input_schema, "output": output_schema}


def _generate_obj_schema(obj):

    if isinstance(obj, pd.core.frame.DataFrame):
        schema = PandasSchemaGenerator.generate_schema(obj)
    elif isinstance(obj, np.ndarray):
        schema = NumpySchemaGenerator.generate_schema(obj)
    else:
        schema = get_swagger_from_python_sample(obj)
    return schema


def get_swagger_from_python_sample(python_data):
    """Generate swagger schema definition for python primitive object."""
    if python_data is None:
        raise ValueError("Python data cannot be None")

    schema = None

    # Python 3 specific type handling
    if type(python_data) is int:
        schema = {"type": "integer", "format": "int64", "example": python_data}
    elif type(python_data) is bytes:
        # Bytes type is not json serializable so will convert to a base 64 string for the sample
        sample = base64.b64encode(python_data).decode("utf-8")
        schema = {"type": "string", "format": "byte", "example": sample}
    elif type(python_data) is range:
        schema = _get_swagger_for_list(
            python_data, {"type": "integer", "format": "int64"}
        )

    # All Python versions handling
    if type(python_data) is str:
        schema = {"type": "string", "example": python_data}
    elif type(python_data) is float:
        schema = {"type": "number", "format": "double", "example": python_data}
    elif type(python_data) is bool:
        schema = {"type": "boolean", "example": python_data}
    elif type(python_data) is datetime.date:
        sample = python_data.strftime("%d-%m-%Y")
        schema = {"type": "string", "format": "date", "example": sample}
    elif type(python_data) is datetime.datetime:
        date_time_with_zone = python_data
        if python_data.tzinfo is None:
            # If no timezone data is passed in, consider UTC
            date_time_with_zone = datetime.datetime(
                python_data.year,
                python_data.month,
                python_data.day,
                python_data.hour,
                python_data.minute,
                python_data.second,
                python_data.microsecond,
                pytz.utc,
            )
        sample = date_time_with_zone.strftime("%d-%m-%Y %H:%M:%S")
        schema = {"type": "string", "format": "date-time", "example": sample}
    elif type(python_data) is datetime.time:
        time_with_zone = python_data
        if python_data.tzinfo is None:
            # If no timezone data is passed in, consider UTC
            time_with_zone = datetime.time(
                python_data.hour,
                python_data.minute,
                python_data.second,
                python_data.microsecond,
                pytz.utc,
            )
        sample = time_with_zone.strftime("%H:%M:%S")
        schema = {"type": "string", "format": "time", "example": sample}
    elif isinstance(python_data, bytearray):
        # Bytes type is not json serializable so will convert to a base 64 string for the sample
        sample = base64.b64encode(python_data).decode("utf-8")
        schema = {"type": "string", "format": "byte", "example": sample}
    elif type(python_data) is list or type(python_data) is tuple:
        schema = _get_swagger_for_list(python_data)
    elif type(python_data) is dict:
        schema = {"type": "dict", "example": python_data}
    # If we didn't match any type yet, try out best to fit this to an object
    if schema is None:
        schema = {"type": "object", "example": python_data}

    # ensure the schema is JSON serializable
    try:
        json.dumps(schema)
    except TypeError as te:
        raise TypeError("ERR_PYTHON_DATA_NOT_JSON_SERIALIZABLE")

    return schema


def _get_swagger_for_list(python_data, item_swagger_type={"type": "object"}):
    sample_size = min(len(python_data), 2)
    sample = []
    for i in range(sample_size):
        sample.append(python_data[i])
    return {"type": "array", "items": item_swagger_type, "example": sample}



# -*- coding: utf-8 -*-
import types
import ast
import pandas as pd
import ipywidgets as widgets
from ipywidgets import interact, interactive, fixed, interact_manual, Layout
from IPython.display import clear_output, display
from fosforml.api import register_model


class RegisterModel:
    def __init__(self):
        self.tab = widgets.Tab()
        self.user_config = list()
        self.tab_names = [
            "Meta Information",
            "Model Information",
            "Data",
            "Optional Features",
        ]

        self.layout_ui_on_screen()

    def layout_ui_on_screen(self):
        self.children = [
            self.meta_information(),
            self.model_information(),
            self.data_information(),
            self.optional_batteries(),
        ]
        self.tab.children = self.children

        self.set_titles()

    def meta_information(self):

        self.model_name = widgets.Text(
            placeholder="(string) name of the model  without space ",
            layout=Layout(width="50%", height="80px"),
        )
        self.model_description = widgets.Textarea(
            placeholder="(string): description of the model",
            layout=Layout(width="50%", height="80px"),
        )

        self.input_type = widgets.RadioButtons(
            position="horizontal",
            options=["json", "csv"],
            value="json",
            layout=Layout(width="30%"),
            description="Input Type",
            disabled=False,
        )

        self.flavor = widgets.Dropdown(
            options=[
                "sklearn",
                "pytorch",
                "keras",
                "tensorflow",
                "pyspark",
                "spacy",
                "application",
                "r",
                "pmml",
                "ensemble",
            ],
            layout=Layout(width="30%"),
            value="sklearn",
            description="Flavor*",
            disabled=False,
        )

        self.model_type = widgets.Dropdown(
            options=[None, "classification", "regression"],
            layout=Layout(width="30%"),
            value=None,
            description="Model Type*",
            disabled=False,
        )

        self.tags = widgets.Text(
            layout=Layout(width="100%"),
            placeholder="Variable Name or (array of strings) user tags associated with the model",
            description="Tags",
        )

        self.init_script = widgets.Textarea(
            description="Init Script", layout=Layout(width="100%", height="40px")
        )

        return widgets.VBox(
            [
                widgets.HBox(
                    [
                        widgets.Label("Model Name*", layout=Layout(width="50%")),
                        widgets.Label("Model Description*", layout=Layout(width="50%")),
                    ]
                ),
                widgets.HBox([self.model_name, self.model_description]),
                widgets.HBox([self.input_type, self.flavor, self.model_type]),
                widgets.HBox([self.init_script]),
                widgets.HBox([self.tags]),
            ]
        )

    def model_information(self):

        self.score_function = widgets.Text(
            placeholder="scoring_func (function): function to be used for scoring",
            layout=Layout(width="50%"),
            disabled=False,
        )
        self.model_object = widgets.Text(
            layout=Layout(width="50%"),
            placeholder="model_obj (object): model to be registered",
            disabled=False,
        )
        self.schema = widgets.Text(
            placeholder="(Dict): input and output schema structure for scoring function",
            layout=Layout(width="100%"),
            disabled=False,
        )
        self.metadata_info = widgets.Text(
            placeholder="metadata_info: metadata information about the version",
            layout=Layout(width="100%"),
            disabled=False,
        )
        self.model_display = widgets.Checkbox(
            description="model_display: If true display model on model list",
            layout=Layout(width="100%"),
        )

        return widgets.VBox(
            [
                widgets.HBox(
                    [
                        widgets.Label("Model Object*", layout=Layout(width="50%")),
                        widgets.Label("Score Function*", layout=Layout(width="50%")),
                    ]
                ),
                widgets.HBox([self.model_object, self.score_function]),
                widgets.HBox(
                    [widgets.Label("Scoring Schema", layout=Layout(width="100%"))]
                ),
                widgets.HBox([self.schema]),
                widgets.Label(
                    "Version Metadata Information", layout=Layout(width="100%")
                ),
                self.metadata_info,
                self.model_display,
            ]
        )

    def data_information(self):
        self.y_true = widgets.Text(description="Y True", placeholder="Var Name")
        self.y_pred = widgets.Text(description="Y Pred", placeholder="Var Name")
        self.prob = widgets.Text(description="Prob", placeholder="Var Name")

        self.features = widgets.Text(
            description="Features", placeholder="features : dummy feature names"
        )
        self.labels = widgets.Text(
            description="Labels", placeholder="labels: predicted labels"
        )

        self.feature_names = widgets.Text(
            description="Feature Names", placeholder="feature_names : all features"
        )
        self.datasource_name = widgets.Text(
            description="Data Source Name", placeholder="datasource_name(string):"
        )

        self.x_train = widgets.Text(
            description="X Train",
            placeholder="x_train (numpy array) : training data of model with feature column",
        )
        self.x_test = widgets.Text(
            description="X Test",
            placeholder="x_test (numpy array) :  test data of model with feature column",
        )

        self.y_train = widgets.Text(
            description="Y Train",
            placeholder="y_train (numpy array) : training data of model with feature column",
        )
        self.y_test = widgets.Text(
            description="Y Test",
            placeholder="y_test (numpy array) :  test data of model with feature column",
        )

        return widgets.VBox(
            [
                widgets.HBox([self.y_true, self.y_pred, self.prob]),
                widgets.HBox([self.labels, self.feature_names, self.datasource_name]),
                widgets.HBox([self.x_train, self.x_test]),
                widgets.HBox([self.y_train, self.y_test]),
            ]
        )

    def optional_batteries(self):
        self.explain_ai = widgets.Checkbox(
            value=False, description="Explain AI", disabled=False
        )

        self.kyd = widgets.Checkbox(
            value=False, description="Know Your Data", disabled=False
        )

        self.kyd_score = widgets.Checkbox(
            value=False, description="Drift Score", disabled=False
        )

        return widgets.VBox([self.explain_ai, self.kyd, self.kyd_score])

    def set_titles(self):
        for i in range(len(self.tab_names)):
            self.tab.set_title(i, self.tab_names[i])

    def variable_inference(self, value):
        if value == "":
            return None, None
        try:
            return_value = eval(value)
        except Exception as e:
            try:
                return_value = ast.literal_eval(value)
            except Exception as e:
                return_value = value

        return type(return_value), return_value

    def build_config(self):

        self.user_config.append(
            dict(parameter_name="name", data_type=str, value=str(self.model_name.value))
        )
        self.user_config.append(
            dict(
                parameter_name="description",
                data_type=str,
                value=str(self.model_description.value),
            )
        )
        self.user_config.append(
            dict(parameter_name="flavour", data_type=str, value=str(self.flavor.value))
        )
        self.user_config.append(
            dict(
                parameter_name="init_script",
                data_type=str,
                value=str(self.init_script.value),
            )
        )

        _list_of_var = [
            (self.tags, "tags"),
            (self.model_type, "model_type"),
            (self.model_object, "model_object"),
            (self.score_function, "score_function"),
            (self.schema, "schema"),
            (self.metadata_info, "metadata_info"),
            (self.model_display, "model_display"),
            (self.y_true, "y_true"),
            (self.y_pred, "y_pred"),
            (self.prob, "prob"),
            (self.labels, "labels"),
            (self.feature_names, "feature_names"),
            (self.datasource_name, "datasource_name"),
            (self.x_train, "x_train"),
            (self.y_train, "y_train"),
            (self.x_test, "x_test"),
            (self.y_test, "y_test"),
            (self.explain_ai, "explain_ai"),
            (self.kyd, "kyd"),
            (self.kyd_score, "kyd_score"),
        ]
        for field, name in _list_of_var:
            data_type, value = self.variable_inference(field.value)
            self.user_config.append(
                dict(parameter_name=name, data_type=data_type, value=value)
            )

        self.user_config_frame = pd.DataFrame(self.user_config)

    def register_model_callback(self, *args, **kwargs):
        self.build_config()
        user_config = dict(
            zip(self.user_config_frame.parameter_name, self.user_config_frame.value)
        )
        clear_output(wait=True)
        display(
            register_model(
                model_obj=user_config.get("model_object"),
                scoring_func=user_config.get("score_function"),
                name=user_config.get("name"),
                description=user_config.get("description"),
                flavour=user_config.get("flavour"),
                tags=user_config.get("tags", None),
                init_script=user_config.get("init_script", None),
                schema=user_config.get("schema", None),
                y_true=user_config.get("y_true", None),
                y_pred=user_config.get("y_pred", None),
                prob=user_config.get("prob", None),
                features=user_config.get("feature_names", None),
                labels=user_config.get("labels", None),
                model_type=user_config.get("model_type", None),
                datasource_name=user_config.get("datasource_name", None),
                metadata_info=user_config.get("metadata_info", None),
                input_type=user_config.get("input_type", None),
                x_train=user_config.get("x_train", None),
                y_train=user_config.get("y_train", None),
                feature_names=user_config.get("feature_names", None),
                explain_ai=user_config.get("explain_ai", False),
                x_test=user_config.get("x_test", None),
                y_test=user_config.get("y_test", None),
                kyd=user_config.get("kyd", False),
                kyd_score=user_config.get("kyd_score", False),
                **kwargs
            )
        )

    def print_all(self, *args, **kwargs):
        self.build_config()

        display(self.user_config_frame)

    def run(self):
        self.css = widgets.HTML(
            "<style >.grad_1{background: #2468a4;} .grad_2{ color:white; background: #2468a4;}</style>"
        )

        self.buttons = widgets.Button(description="Display")
        self.buttons.on_click(self.print_all)
        self.register_button = widgets.Button(description="Register")
        self.register_button.on_click(self.register_model_callback)

        self.logo = widgets.HTML(value="<h3>Model Registeration</h3>")
        self.logo.add_class("grad_2")
        self.dashboard = widgets.VBox(
            [
                self.css,
                self.logo,
                self.tab,
                widgets.HBox([self.buttons, self.register_button]),
            ]
        )
        self.dashboard.add_class("grad_1")
        display(self.dashboard)



import ipywidgets as widgets
from ipywidgets import interact, interactive, fixed, interact_manual, Layout


class Fields(object):
    def text(self, description="", placeholder="", width="50%", disabled=True, value=None):
        return widgets.Text(
            description=description,
            placeholder=placeholder,
            layout=Layout(width=width),
            disabled=disabled,
            value=value
        )

    def textarea(self, description="", placeholder="", width="50%", disabled=True, value=None, *args, **kwargs):
        return widgets.Textarea(
            description=description,
            placeholder=placeholder,
            layout=Layout(width=width),
            disabled=disabled,
            value=value,
            *args,
            **kwargs
        )

    def radio(self, description="", options=[], placeholder="", width="50%", disabled=True, value=None, *args,
              **kwargs):
        return widgets.RadioButtons(
            options=options,
            description=description,
            placeholder=placeholder,
            layout=Layout(width=width),
            disabled=disabled,
            value=value,
            *args,
            **kwargs
        )

    def label(self, label, layout="50%"):
        return widgets.Label(label, layout=Layout(width=layout))

    def datetime(self, description, value, disabled=False):
        return widgets.DatePicker(
            description=description,
            disabled=disabled,
            value=value
        )

    def marks(self, description, value):
        return widgets.Valid(
            value=value,
            description=description,
        )

    def checkbox(self, description, value, disabled=True, indent=False):
        return widgets.Checkbox(
            value=value,
            description=description,
            disabled=disabled,
            indent=indent
        )

    def select_multiple(self, description, options, disabled=True, value=None, rows=10):
        return widgets.SelectMultiple(
            options=options,
            value=value,
            rows=rows,
            description=description,
            disabled=disabled
        )

    def button(self, description, icon):
        return widgets.Button(description=description, icon=icon)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function
from datetime import datetime

import numpy as np

from ..schema.base_schema_generator import BaseSchemaGenerator


class NumpySchemaGenerator(BaseSchemaGenerator):
    """Schema generator for numpy ndarray."""

    @classmethod
    def generate_schema(cls, ndarray):
        swagger_item_type = cls.get_swagger_for_dtype(ndarray.dtype)
        swagger_schema = cls.get_swagger_for_array_type(
            swagger_item_type, ndarray.shape
        )

        # Create a couple of example records
        count = min(len(ndarray), 2)
        examples = cls.get_swagger_examples(ndarray, swagger_schema["items"], count)
        swagger_schema["example"] = examples
        return swagger_schema

    @classmethod
    def _date_item_to_string(cls, date_item):
        return date_item.astype(datetime).strftime("%Y-%m-%d")

    @classmethod
    def _timestamp_item_to_string(cls, date_item):
        return date_item.astype(datetime).strftime("%Y-%m-%d %H:%M:%S,%f")



# -*- coding: utf-8 -*-


class ConfigError(Exception):
    pass


class MosaicException(Exception):
    """Mosaic ai common exception"""

    code = 500
    message = "Something went wrong"


class InvalidModelIDException(MosaicException):
    """Mosaic ai authentication error"""

    code = 500
    message = "The Model ID provided is invalid. Kindly provide a valid Model ID !"



# -*- coding: utf-8 -*-
# Pretty Prints Model Describe
import pprint

import ipywidgets as widgets
from IPython.display import clear_output, display
from dateutil import parser
from fosforml.widgets.fields import Fields


class ModelDescribe(Fields):
    def __init__(self, ml_model_id=None):
        self.pp = pprint.PrettyPrinter(indent=2)
        self.ml_model_id = ml_model_id
        self._data = {}

        self.datetime_format = '%Y-%m-%dT%H:%M:%S%z'

        self.tab = widgets.Tab()
        self.tab_names = [
            "Model Metadata",
            "Version Metadata",
        ]

    def set_titles(self):
        for i in range(len(self.tab_names)):
            self.tab.set_title(i, self.tab_names[i])

    def model_meta_information(self):
        self.modelid_widget_ = self.text("Model id", value=self._data.get("id", ""), disabled=False)

        return widgets.VBox(
            [
                widgets.HBox(
                    [
                        self.modelid_widget_,
                        self.text("Name", value=self._data.get("name", "")),
                        self.text("Flavour", value=self._data.get("flavour", "")),
                    ]
                ),
                widgets.HBox(
                    [
                        self.textarea("Description", value=self._data.get("description", False)),
                        self.text("Project", value=self._data.get("project_id", "")),
                    ]
                ),
                widgets.HBox(
                    [
                        self.checkbox("Deployment Status", value=self._data.get("deploymentstatus", False)),
                        self.checkbox("Model Display", value=self._data.get("model_display", False))
                    ]
                ),
                widgets.HBox(
                    [
                        self.text("Source", value=self._data.get("source", "")),
                        self.text("Tag", value=self._data.get("tags", "")),
                        self.text("Type", value=self._data.get("type", "")),
                    ]
                ),
                widgets.HBox(
                    [
                        self.text("Created By", "25%", value=self._data.get("created_by", "")),
                        self.datetime("On",
                                      value=parser.parse(self._data.get("created_on"))),
                    ]
                ),
                widgets.HBox(
                    [
                        self.text("Modified By", "25%", value=self._data.get("last_modified_by", "")),
                        self.datetime("On", value=parser.parse(self._data.get("last_modified_on"))),
                    ]
                ),

            ]
        )

    def init_script_dashboard(self, init_script):
        script_ = str(init_script).replace('\\n', '\n').replace('\\t', '\t').replace('\"', "")
        init_script_dashboard = widgets.VBox([
            widgets.HBox([self.textarea("Init Script", value=script_, width="100%", rows=20)]),
        ])
        return init_script_dashboard

    def model_info_dashboard(self, model_info):
        if not model_info:
            return widgets.HBox([widgets.Label("Model Info Is Not Present")])

        features_name = model_info.get("features_name", "")
        if isinstance(features_name, str) or not features_name:
            features_name = ["Empty"]

        model_info_dashboard = widgets.VBox([
            widgets.HBox([self.checkbox("Deeplearning Model", value=model_info.get("deep_learning_model", False)),
                          self.checkbox("Know Your Data", value=model_info.get("kyd", False)),
                          self.checkbox("Explain AI", value=model_info.get("expai", False)),
                          self.checkbox("Infer Feature DType", value=model_info.get("feature_type_inferenced", False)),
                          ]),
            widgets.HBox([self.select_multiple("Features",
                                               features_name,
                                               value=features_name),
                          self.textarea("Feature Type", value=self.pp.pformat(model_info.get("features_type", "")),
                                        rows=8)
                          ]),
            widgets.HBox(
                [self.radio("Mode", options=["classification", "regression"], value=model_info.get("mode", "")),
                 self.text("No. Features", value=str(model_info.get("number_of_features", ""))),
                 self.text("No. Targets", value=str(model_info.get("number_of_targets", "")))
                 ]),
            widgets.HBox(
                [self.textarea("Target Mapping", value=self.pp.pformat(model_info.get("targets_mapping", "")), rows=8),
                 self.textarea("Temp Dir", value=str(model_info.get("temp_dir", "")), rows=8)
                 ]),

        ])

        return model_info_dashboard

    def model_class_dashboard(self, model_class):
        model_class_ = str(model_class).replace('\\n', '\n').replace('\\t', '\t').replace('\"', "")
        model_class_dashboard = widgets.VBox([
            widgets.HBox([self.textarea("Model Class", value=model_class_, width="100%", rows=20)]),
        ])
        return model_class_dashboard

    def deploy_info_dashboard(self, deploy_info):
        deploy_info_dashboard = widgets.VBox([
            widgets.HBox(
                [self.textarea("Deploy Info", value=str(deploy_info.get("deploy_info", "")), width="100%", rows=5)]),
            widgets.HBox(
                [self.textarea("Deployment", value=str(deploy_info.get("deployments", "")), width="100%", rows=5)]),
        ])
        return deploy_info_dashboard

    def others_dashboard(self, version_info):

        others_info_dashboard = widgets.VBox([
            widgets.HBox(
                [self.textarea("Profiling", value=str(version_info.get("profiling", "")), width="100%", rows=5)]),
            widgets.HBox([self.textarea("Schema", value=str(version_info.get("schema", "")), width="100%", rows=5)]),
        ])
        return others_info_dashboard

    def version_meta_information(self):
        children_ = []
        titles_ = []
        for each_version in self._data["versions"]:
            detail_tab = widgets.Tab()
            detail_tab_children = []
            detail_titles_ = ["Image", "Model Information", "Init Script", "Model Class", "Deploy Info", "Others"]

            image_dashboard = widgets.VBox([
                self.text("CPU", value=str(each_version.get("docker_image_url", "")), width="80%"),
                self.text("GPU", value=str(each_version.get("gpu_docker_image_url", "")), width="80%"),

            ])

            detail_tab_children.append(image_dashboard)

            model_info_dashboard = self.model_info_dashboard(each_version.get("model_info", ""))
            detail_tab_children.append(model_info_dashboard)

            init_script_dashboard = self.init_script_dashboard(each_version.get("init_script", ""))
            detail_tab_children.append(init_script_dashboard)

            model_class_dashboard = self.model_class_dashboard(each_version.get("model_class", ""))
            detail_tab_children.append(model_class_dashboard)

            deploy_info_dashboard = self.deploy_info_dashboard(each_version)
            detail_tab_children.append(deploy_info_dashboard)

            others_dashboard = self.others_dashboard(each_version)
            detail_tab_children.append(others_dashboard)

            detail_tab.children = detail_tab_children
            for i in range(len(detail_tab_children)):
                detail_tab.set_title(i, detail_titles_[i])

            dashboard = widgets.VBox([
                widgets.HBox([self.text("No", value=str(each_version.get("version_no", ""))),
                              self.text("Id", value=str(each_version.get("id", ""))),
                              self.text("Status", value=str(each_version.get("description", ""))),
                              ]),
                widgets.HBox([self.text("Datasource", value=str(each_version.get("datasource_name", ""))),
                              self.text("Deployment Model", value=str(each_version.get("dependent_model", ""))),
                              self.textarea("Description", value=str(each_version.get("description", ""))),
                              ]),
                detail_tab,
                widgets.HBox([self.text("Input Type", value=str(each_version.get("input_type", ""))),
                              self.text("Model Id", value=str(each_version.get("ml_model_id", ""))),

                              ]),
                widgets.HBox([self.text("Object URL", value=str(each_version.get("object_url", "")), width="100%"),
                              ]),
                widgets.HBox(
                    [
                        self.text("Created By", "25%", value=each_version.get("created_by", "")),
                        self.datetime("On",
                                      value=parser.parse(each_version.get("created_on", ""))),
                    ]
                ),
                widgets.HBox(
                    [
                        self.text("Modified By", "25%", value=each_version.get("last_modified_by", "")),
                        self.datetime("On", value=parser.parse(each_version.get("last_modified_on", ""))),
                    ]
                ),
            ])

            children_.append(dashboard)
            titles_.append(str(each_version.get("version_no", "")) + ") " + each_version.get("id", ""))
        accordion = widgets.Accordion(children=children_, titles=titles_)
        for i in range(len(children_)):
            accordion.set_title(i, titles_[i])
        return accordion

    def print_raw(self, *args, **kwargs):
        return display(self._data)

    def clear_raw(self, *args, **kwargs):
        clear_output(wait=True)

    def describe_button_callback(self, *args, **kwargs):
        from fosforml.api import describe_model
        self._data = describe_model(self.modelid_widget_.value)
        self.clear_raw()
        self.view(load_data=False)

    def view(self, use_data=None, load_data=True):
        try:
            if load_data:
                from fosforml.api import describe_model
                self._data = describe_model(self.ml_model_id)

            if use_data:
                self._data = use_data

            self.children = [
                self.model_meta_information(),
                self.version_meta_information()
            ]
            self.tab.children = self.children

            self.set_titles()

            self.css = widgets.HTML(
                "<style>.grad_1{background: #2468a4;} .grad_2{ color:white; background: #2468a4;}</style>"
            )

            self.buttons = self.button("Print Raw", 'eye')
            self.buttons.on_click(self.print_raw)
            self.clear_button = self.button("Close", 'close')
            self.clear_button.on_click(self.clear_raw)
            self.describe_button = self.button("Describe", 'binoculars')
            self.describe_button.on_click(self.describe_button_callback)

            self.logo = widgets.HTML(value="<h3>Model Describe</h3>")
            self.logo.add_class("grad_2")
            self.dashboard = widgets.VBox(
                [
                    self.css,
                    self.logo,
                    self.tab,
                    widgets.HBox([self.buttons, self.clear_button, self.describe_button]),
                ]
            )
            self.dashboard.add_class("grad_1")
            display(self.dashboard)
        except Exception as e:
            print(self._data)



# -*- coding: utf-8 -*-
import inspect


class scoring_func(object):
    """decorator for defining user defined scoring function"""

    def __init__(self, user_func):
        # make sure it is a function
        if not inspect.isfunction(user_func):
            raise TypeError("Argument to scoring_func decorator must be a function")
        # make sure it accepts two parameters
        if user_func.__code__.co_argcount != 2:
            raise TypeError("User defined scoring function should accept two arguments")
        self._user_func = user_func

    def __call__(self, *args, **kwargs):
        return self._user_func(*args, **kwargs)


# -*- coding: utf-8 -*-

from __future__ import absolute_import
from datetime import datetime

import numpy as np
import pandas as pd

from ..schema.base_schema_generator import BaseSchemaGenerator


class PandasSchemaGenerator(BaseSchemaGenerator):
    """Schema generator for pandas dataframe."""

    @classmethod
    def generate_schema(cls, df):
        cnames = df.columns.values.tolist()
        dtypes = df.dtypes.tolist()
        properties = {}
        for cname, dtype in zip(cnames, dtypes):
            if dtype.name == "object" and type(df[cname][0]) is str:
                dtype = np.dtype("str")
            properties[cname] = cls.get_swagger_for_dtype(dtype)

        record_schema = {"type": "object", "properties": properties}
        # Create a couple of example records
        count = min(len(df), 2)
        examples = cls.get_swagger_examples(df.iloc, record_schema, count)
        schema = {
            "type": "pd.core.frame.DataFrame",
            "items": record_schema,
            "example": examples,
        }
        return schema

    @classmethod
    def _date_item_to_string(cls, date_item):
        return date_item.astype(datetime).strftime("%Y-%m-%d")

    @classmethod
    def _timestamp_item_to_string(cls, date_item):
        return pd.Timestamp(date_item).to_pydatetime().strftime("%Y-%m-%d %H:%M:%S,%f")



# -*- coding: utf-8 -*-

from typing import Any
import os,requests,json
from fosforml.constants import connection_manager_url, UserConfigs

class snowflakesession:
    def __init__(self):
        self.connection_params = None
        self.session = None
        self._connection_details()

    def _connection_details(self):
        try:
            if UserConfigs.connection_details in os.environ :
                connection_info = os.environ.get(UserConfigs.connection_details)
                self.connection_params = json.loads(connection_info.replace("\\", ""))

            elif not self.connection_params and  not "connection_details" in os.environ:
                project_id = os.getenv(UserConfigs.project_id)

                if not project_id:
                    raise ValueError(f"{UserConfigs.project_id} not set")
                
                if not connection_manager_url:
                    raise ValueError("CONNECTION_MANAGER_BASE_URL not set")
                
                url = f"{connection_manager_url}/connections/api/ConnectionManager/v1/allConnections?projectId={project_id}"
                
                connection_details = requests.get(url, verify=False).json()[0]["connectionDetails"]
                self.connection_params = {
                    'account':f"{connection_details['accountName']}.{connection_details['region']}.{connection_details['cloudPlatform']}",
                    "user": connection_details["dbUserName"],
                    "password": connection_details["dbPassword"],
                    "database": connection_details["defaultDb"],
                    "schema": connection_details["defaultSchema"],
                    "warehouse": connection_details["wareHouse"],
                    "role": connection_details["role"],
                }

        except Exception as e:
            raise Exception(f"Failed to get connection details. {str(e)}")
        
    def get_session(self):
        if not self.session:
            try:
                connection_parameters = self.validate_connection_params(self.connection_params)
                ## snowflake session creation                
                from snowflake.snowpark import Session
                self.session = Session.builder.configs(connection_parameters).create()
            except Exception as e:
                raise Exception(f"Failed to create snowflake session. {str(e)}")
            
        return self.session
    
    @staticmethod
    def validate_connection_params(connection_params):
        if not connection_params["account"]:
            raise ValueError("Account name is required")
        if not connection_params["user"]:
            raise ValueError("Username is required")
        if not connection_params["password"]:
            raise ValueError("Password is required")
        if not connection_params["database"]:
            raise ValueError("Database is required")
        if not connection_params["schema"]:
            raise ValueError("Schema is required")
        if not connection_params["warehouse"]: 
            raise ValueError("Warehouse is required")
        if not connection_params["role"]:
            raise ValueError("Role is required")
        
        return connection_params
    
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        if not self.session:
            return self.get_session()
        
        return self.session

    def __enter__(self):
        if not self.session:
            return self.get_session()
        return self.session
    
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            raise Exception(exc_type, exc_value, traceback)
        
        if self.session:
            self.session.close()

    def execute(self, query):
        pass

    def close(self):
        pass


def get_session():
    try:
        session_obj = snowflakesession()
        return session_obj.get_session()
    except Exception as msg:
        return f"Failed to create snowflake session. {msg}"
    




# -*- coding: utf-8 -*-
from __future__ import absolute_import
import base64
from abc import ABCMeta, abstractmethod

from six import with_metaclass


class BaseSchemaGenerator(with_metaclass(ABCMeta)):
    """Abstract base class for Swagger scheme generatorself.

    TODO....
    """

    @classmethod
    @abstractmethod
    def generate_schema(cls, obj):
        """Must be implemented in sub-class."""
        pass

    _converters = {
        "float16": {"type": "number", "format": "float"},
        "float32": {"type": "number", "format": "float"},
        "float64": {"type": "number", "format": "double"},
        "int8": {"type": "integer", "format": "int8"},
        "int16": {"type": "integer", "format": "int16"},
        "int32": {"type": "integer", "format": "int32"},
        "int64": {"type": "integer", "format": "int64"},
        "uint8": {"type": "integer", "format": "uint8"},
        "uint16": {"type": "integer", "format": "uint16"},
        "uint32": {"type": "integer", "format": "uint32"},
        "uint64": {"type": "integer", "format": "uint64"},
        "bool": {"type": "boolean"},
        "object": {"type": "object"},
    }

    @classmethod
    def get_swagger_for_dtype(cls, dtype):
        if len(dtype) == 0:
            if dtype.subdtype is None:  # scalar type
                swag = cls._get_swagger_for_scalar_dtype(dtype)
            else:  # sub-array type
                swag_subtype = cls.get_swagger_for_dtype(dtype.subdtype[0])
                swag = cls.get_swagger_for_array_type(swag_subtype, dtype.subdtype[1])
        else:  # structured type
            properties = {
                name: cls.get_swagger_for_dtype(dtype[name]) for name in dtype.names
            }
            swag = {"type": "object", "properties": properties}
        return swag

    @classmethod
    def get_swagger_for_array_type(cls, item_swag_type, shape):
        swag = {"type": "array", "items": item_swag_type}
        if len(shape) > 1:
            for _ in range(len(shape) - 1):
                swag = {"type": "np.ndarray", "items": swag}
        return swag

    @classmethod
    def _get_swagger_for_scalar_dtype(cls, dtype):
        name = dtype.name.lower()
        if name in cls._converters:
            return cls._converters.get(name)
        elif name.startswith("datetime"):
            return {"type": "string", "format": "date-time"}
        elif name.startswith("str"):
            return {"type": "string"}
        elif name.startswith("bytes") or type_name.startswith("void"):
            return {"type": "string", "format": "binary"}
        elif name.startswith("timedelta"):
            return {"type": "string", "format": "timedelta"}
        else:
            raise TypeError("Unsupported data type '{}'".format(name))

    @classmethod
    def get_swagger_examples(cls, iterable, record_schema, count):
        examples = []
        for i in range(count):
            example = cls._get_swagger_example_record(record_schema, iterable[i])
            examples.append(example)
        return examples

    @classmethod
    def _get_swagger_example_record(cls, item_swagger_schema, data_item):
        item_type = item_swagger_schema["type"]
        if item_type == "object":
            if "properties" in item_swagger_schema:
                sample_swag = dict()
                for field in item_swagger_schema["properties"]:
                    sample_swag[field] = cls._get_swagger_example_record(
                        item_swagger_schema["properties"][field], data_item[field]
                    )
            elif "additionalProperties" in item_swagger_schema:
                sample_swag = dict()
                for field in data_item:
                    sample_swag[field] = cls._get_swagger_example_record(
                        item_swagger_schema["additionalProperties"], data_item[field]
                    )
            else:
                sample_swag = str(data_item)
        elif item_swagger_schema["type"] == "array":
            sample_swag = []
            subarray_item_swagger = item_swagger_schema["items"]
            for i in range(len(data_item)):
                array_item_sample = cls._get_swagger_example_record(
                    subarray_item_swagger, data_item[i]
                )
                sample_swag.append(array_item_sample)
        elif item_type == "number":
            sample_swag = float(data_item)
        elif item_type == "integer":
            sample_swag = int(data_item)
        elif item_type == "bool":
            sample_swag = bool(data_item)
        elif item_type == "string" and "format" in item_swagger_schema:
            if item_swagger_schema["format"] == "date":
                sample_swag = cls._date_item_to_string(data_item)
            elif item_swagger_schema["format"] == "date-time":
                sample_swag = cls._timestamp_item_to_string(data_item)
            elif item_swagger_schema["format"] == "binary":
                sample_swag = base64.b64encode(data_item).decode("utf-8")
            else:
                sample_swag = str(data_item)
        else:
            sample_swag = str(data_item)
        return sample_swag

    @classmethod
    @abstractmethod
    def _date_item_to_string(cls, date_item):
        pass

    @classmethod
    @abstractmethod
    def _timestamp_item_to_string(cls, date_item):
        pass



from snowflake.ml.registry import Registry
from .snowflakesession import snowflakesession as SnowflakeSession
from .utilities import DatasetManager, Metadata
from fosforml.utils import get_headers
from fosforml.constants import NotebooksAPI
from datetime import datetime
import json,requests,os
import pandas as pd
from fosforml.constants import ModelConstants
from fosforml.model_manager import snowflakesession

class ModelRegistry:
    def __init__(self,session):
        self.session_instance = SnowflakeSession()
        self.snowflake_session = session
        self.model_registry = Registry(session=self.snowflake_session, 
                                       database_name=self.session_instance.connection_params['database'],
                                       schema_name=self.session_instance.connection_params['schema']
                                       )
        self.models_info = self.model_registry.show_models()

    def register_model(self,
                       model,
                       score,
                       model_name,
                       dataset_name,
                       dataset_source,
                       description,
                       conda_dependencies,
                       model_flavour,
                       model_type,
                       sf_input_dataframe,
                       x_train,
                       y_train,
                       x_test,
                       y_test,
                       y_pred,
                       prob,
                       python_version,
                       source,
                       metadata,
                       prediction_column=None,
                       target_column=None
                       ):
        
        if score :
            return False, "score fuction not supported, implementation is in progress ..."

        versioning_errors, model_version = self.get_model_version(self.model_registry,model_name)
        if versioning_errors:
            return False,model_version
        
        metrics_errors,metrics = self.get_model_metrics(
                                        source=source,
                                        model=model,
                                        session=self.snowflake_session,
                                        metadata=metadata,
                                        model_type=model_type,
                                        model_flavour=model_flavour,
                                        x_train=x_train,
                                        y_train=y_train,
                                        x_test=x_test,
                                        y_test=y_test,
                                        y_pred=y_pred,
                                        prob=prob,
                                        sf_input_dataframe=sf_input_dataframe,
                                        dataset_source=dataset_source,
                                        dataset_name=dataset_name,
                                        prediction_column=prediction_column,
                                        target_column=target_column
                                        )
        if metrics_errors:
            return False,metrics

        metrics['python_version'] = python_version
        metrics['conda_dependencies'] = conda_dependencies
        metrics['algorithm_type'] = os.getenv('algorithm_type',None)
        metrics['algorithm'] = os.getenv('algorithm',None)
        metrics['experiment_name'] = os.getenv('experiment_name',None)
        metrics['run_id'] = os.getenv('run_id',None)
        metrics['experiment_id'] = os.getenv('experiment_id',None)
        metrics['flavour'] = model_flavour
        metrics['model_type'] = model_type

        ## register model

        try:
            self.model_registry.log_model(
                   model,
                   model_name=model_name,
                   version_name=model_version,
                   comment=description,
                   conda_dependencies=conda_dependencies,
                   metrics=metrics,
                   sample_input_data=sf_input_dataframe if model_flavour.lower() == "snowflake" else x_test,
                   python_version=python_version  
                  )
        except Exception as e:
            return False,f"Failed to register model '{model_name}'. {str(e)}"
        
        ## upload model datasets
        dataset_manager = DatasetManager(model_name=model_name,
                                         version_name=model_version,
                                         session=self.snowflake_session
                                         )
        ds_status,ds_message = dataset_manager.upload_datasets(
                                session=self.snowflake_session,
                                datasets={
                                        # "x_train": x_train,
                                        # "y_train": y_train,
                                        "x_test": x_test,
                                        "y_test": y_test,
                                        # "y_pred": y_pred,
                                        # "prob": prob
                                    })

        if not ds_status:
            return False,ds_message

        ## update model metadata
        metadata = Metadata(model_registry=self.model_registry)
        metadata_status,metadata_message = metadata.update_model_registry(
            session=self.snowflake_session,
            model_name=model_name,
            model_description=description,
            model_tags={
                "FLAVOR": model_flavour,
                "MODELTYPE": model_type,
                "CREATEDON" : metrics['created_on']
            }
        )

        if not metadata_status:
            return False,metadata_message
        
        return True,f"Model '{model_name}' registered successfully."

    def update_model_details(self,
                             model_name,
                             comments,
                             model_tags):
        metadata = Metadata(model_registry=self.model_registry)
        return metadata.update_model_registry(
            session=self.snowflake_session,
            model_name=model_name,
            model_description=comments,
            model_tags=model_tags
        )

    def get_model_details(self, model_name):
        pass

    def get_model_version(self, registry, model_name):
        model_info = pd.DataFrame(); models_data = pd.DataFrame()
        try:
            models_data = registry.show_models()       
        except Exception as e:
            return True,f"Failed to get model version. {str(e)}"
        
        if models_data.empty :
            return False,"v1"
                
        model_info = models_data[models_data['name']==model_name.upper()]

        if model_info.empty:
            return False,"v1"

        model_info = models_data[models_data['name']==model_name.upper()]
        model_versions = json.loads(model_info.versions.to_list()[0])
        if not model_versions:
            return False,"v1"
        else:
            last_version = max(model_versions,key=lambda x: int(x[1:]),default='v0')
            return False, f"v{int(last_version[1:])+1}"
    
    def get_model_metrics(self,
                          source,
                          model,
                          session,
                          metadata,
                          model_type,
                          model_flavour,
                          x_train=None,
                          y_train=None,
                          x_test=None,
                          y_test=None,
                          y_pred=None,
                          prob=None,
                          sf_input_dataframe=None,
                          dataset_source="",
                          dataset_name="",
                          prediction_column=None,
                          target_column=None
                            ):
        
        metrics = {}
        dataset_errors,model_dataset_details = self.get_dataset_details(
            model=model,session=session,
            model_flavour=model_flavour,
            x_train=x_train,
            y_train=y_train,
            x_test=x_test,
            y_test=y_test,
            y_pred=y_pred,
            prob=prob,
            sf_input_dataframe=sf_input_dataframe,
            dataset_name=dataset_name,
            dataset_source=dataset_source,
            target_column=target_column,
            prediction_column=prediction_column,
            model_type=model_type
        )
        if dataset_errors:
            return True, model_dataset_details['message']
        
        metrics['dataset_details'] = model_dataset_details['ui_display'] if 'ui_display' in model_dataset_details else {}

        buildtime_metrics_errors,buildtime_metrics = self.get_model_performance_metrics(model,
                                                                      model_flavour,
                                                                      model_type,
                                                                      model_dataset_details
                                                                      )
        if buildtime_metrics_errors:
            return True,buildtime_metrics['message']
        
        metrics['model_metrics'] = buildtime_metrics

        metrics['hyper_parameters'] = model_dataset_details['hyper_parameters']

        source_errors,source_info = self.get_source_details(source)
        if source_errors:
            return True,source_info['message']

        metrics['created_on'] = session.sql("SELECT TO_CHAR(CURRENT_TIMESTAMP,'YYYY-MM-DD HH24:MI:SS')").to_pandas().squeeze()
        metrics['created_by'] = os.getenv('userId',None)
        metrics['status'] = source_info['status']
        metrics['source'] = source_info['source']
        if metrics['source'] == "NOTEBOOK":
            metrics['Repo_Details'] = source_info['repo_details']
        metrics['metadata'] = metadata if metadata else None

        return False, metrics

    def get_hyper_parameters(self,model):
        try:
            if str(type(model)).find("snowflake") > 0:
                return model.to_sklearn().get_params()
            else:
                return model.get_params()
        except Exception as e:
            return f"Failed to get hyper parameters. {str(e)}"

    def get_model_performance_metrics(self,model,
                                      model_flavour,
                                      model_type,
                                      model_dataset_details
                                      ):
        try:
            sf_true_cn = [] ; sf_pred_cn = [] ; sf_pred_proba_cn = [] 
            sf_input_cols = model_dataset_details['metadata']['feature_names']
            if model_flavour.lower() == ModelConstants.sklearn_model_flavour and str(type(model)).find(ModelConstants.sklearn_model_flavour) > 0:
                

                for col in model_dataset_details['final_df'].columns:
                    if col[1:-1].lower() == model_dataset_details['metadata']['target_column'][0].lower():
                        sf_true_cn.append(col)

                    if col[1:-1].lower() == model_dataset_details['metadata']['prediction_column'][0].lower():
                        sf_pred_cn.append(col)
                    
                    if col in model_dataset_details['metadata']['prob_columns']:
                        sf_pred_proba_cn.append(col)
                    else:
                        for prob_col in model_dataset_details['metadata']['prob_columns']:
                            if prob_col in col:
                                sf_pred_proba_cn.append(col)
                                
                                            
            else:
                sf_true_cn = model_dataset_details['metadata']['target_column']
                sf_pred_cn = model_dataset_details['metadata']['prediction_column']
                sf_pred_proba_cn = model_dataset_details['metadata']['prob_columns']

            if sf_true_cn is None or sf_pred_cn is None :
                return True, {"message": "Failed to get 'target_column' and 'prediction_column'."}

            if model_type.lower() == "classification":
                from fosforml.model_manager.model_metrics import Classification
                classification_metrics = Classification(
                    model_obj=model_dataset_details['model_obj'],
                    sf_df=model_dataset_details['final_df'],
                    true_cn=sf_true_cn,
                    pred_cn=sf_pred_cn,
                    pred_proba_cn=sf_pred_proba_cn,
                    sf_input_cols=sf_input_cols
                )
                return False, classification_metrics.get_metrics()
            
            if model_type.lower() == "regression":
                from fosforml.model_manager.model_metrics import Regression
                regression_metrics = Regression(
                                                model=model_dataset_details['model_obj'],
                                                sf_df=model_dataset_details['final_df'], 
                                                true_cn=sf_true_cn,
                                                pred_cn=sf_pred_cn,
                                                pred_proba_cn=sf_pred_proba_cn,
                                                sf_input_cols=sf_input_cols
                                                )
                return False, regression_metrics.get_metrics()
            
            else:
                return True,{"message": "Invalid model type, please provide valid model type as classification or regression"} 
            
        except Exception as msg:
            raise Exception(f"Failed to get model performance metrics. {str(msg)}")
            # return True, {"message": f"Failed to get model performance metrics. {str(msg)}"}
        
        # dummy_data = [{'tag': 'feature_importance', 'model_metric_value': [{'column_name': 'DEPARTMENT', 'importance': 0.14061672606947828}, {'column_name': 'SATISFACTION_LEVEL', 'importance': 0.1104329386548882}, {'column_name': 'LAST_EVALUATION', 'importance': 0.1350624049528422}, {'column_name': 'NUMBER_PROJECT', 'importance': 0.04248402957988852}, {'column_name': 'AVERAGE_MONTLY_HOURS', 'importance': 0.2118735496815215}, {'column_name': 'TIME_SPEND_COMPANY', 'importance': 0.10658202765565672}, {'column_name': 'WORK_ACCIDENT', 'importance': 0.010641980974635002}, {'column_name': 'LEFT', 'importance': 0.16529938539291036}, {'column_name': 'PROMOTION_LAST_5YEARS', 'importance': 0.0770069570381792}]}, {'tag': 'confusion_matrix', 'model_metric_value': [{'column_1_counter': 0, 'column_2_counter': 0, 'prediction': 1404.0, 'column_1': 0, 'column_2': 0}, {'column_1_counter': 0, 'column_2_counter': 1, 'prediction': 380.0, 'column_1': 0, 'column_2': 1}, {'column_1_counter': 1, 'column_2_counter': 0, 'prediction': 1129.0, 'column_1': 1, 'column_2': 0}, {'column_1_counter': 1, 'column_2_counter': 1, 'prediction': 483.0, 'column_1': 1, 'column_2': 1}]}, {'tag': 'detailed_matrix', 'model_metric_value': {'accuracy_score': 0.555654, 'precision_score': 0.5596755504055619, 'recall_score': 0.29962779156327546, 'f1_score': 0.3903030303030303, 'log_loss': 16.01586365258474, 'roc_auc_score': 0.5433116536291713}}, {'tag': 'roc_auc', 'model_metric_value': {'fpr': [0.0, 0.21300448430493274, 1.0], 'tpr': [0.0, 0.29962779156327546, 1.0], 'data': 0.5433116536291713}}]
        # return "dummy_data"
    
    def get_source_details(self,source):
        source_info = {}
        if source is None or source == "":
            source_info['status'] = "Registered"
            source_info['source'] = "NOTEBOOK"
            source_info['repo_details'] = self.get_repo_details()

        elif source is not None and  source.upper() == "EXPERIMENT":
            source_info['status'] = "Experimented"
            source_info['source'] = "EXPERIMENT"

        elif source is not None and source.upper() == "BYOM":
            source_info['status'] = "Deploying"
            source_info['source'] = "BYOM"
            
        elif source is not None and source.upper() == "NOTEBOOK" :
            source_info['status'] = "Registered"
            source_info['source'] = "NOTEBOOK"
            source_info['repo_details'] = self.get_repo_details()    
        else:
            return True, {"message": f"Invalid source. Please provide valid source as Experiment, BYOM or Notebook."}

        return False,source_info

    def get_repo_details(self):
        from fosforml.constants import UserConfigs
        repo_key = UserConfigs.repo_key
        repo_details = []
        
        try:
            if repo_key.lower() in os.environ or repo_key.upper() in os.environ:
                repo_details = os.environ.get(repo_key,'[]')
                return json.loads(repo_details.replace("\\",""))
            else:
                url = NotebooksAPI.notebooks_api_server_url + NotebooksAPI.git_repo
                headers = get_headers()
                response = requests.get(url, headers=headers).json()
                return response
        except Exception as e:
            return repo_details

    def get_dataset_details(self,model,
            session,
            model_flavour,
            x_train=None,
            y_train=None,
            x_test=None,
            y_test=None,
            y_pred=None,
            prob=None,
            sf_input_dataframe=None,
            dataset_name="",
            dataset_source="",
            target_column=None,
            prediction_column=None,
            model_type=None):
        

        dataset_details = {}
        dataset_metadata = {}
        dataset_ui_display = []
        purpose = DatasetManager.get_dataset_purpose
        connection_params = snowflakesession().connection_params
        if model_flavour.lower() == "snowflake" and str(type(model)).find("snowflake") > 0:
            try:

                if not hasattr(model,"label_cols") and target_column is None:
                   return True, {"message": f"Failed to register model, 'target_column' is required to register the model ."}
                
                if not hasattr(model,"output_cols") and prediction_column is None:
                   return True, {"message": f"Failed to register model, 'prediction_column' is required to register the model ."}
                
                if len(model.label_cols) < 1 and target_column is None:
                   return True, {"message": f"Failed to register model, 'target_column' is required to register the model ."}
                
                if len(model.output_cols) < 1 and prediction_column is None:
                   return True, {"message": f"Failed to register model, 'prediction_column' is required to register the model ."}


                dataset_metadata['feature_names'] = model.input_cols if hasattr(model,"input_cols") else []
                dataset_metadata['target_column'] = model.label_cols if all([hasattr(model,"label_cols"),len(model.label_cols)>0]) else [target_column]
                dataset_metadata['prediction_column'] = model.output_cols if all([hasattr(model,"output_cols"),len(model.output_cols)>0]) else [prediction_column]

                from fosforml.utils import load_model_artifacts
                try:
                    model_artifacts = load_model_artifacts(
                                                        session=session,
                                                        model_type=model_type,
                                                        model_obj=model,
                                                        snowflake_model=True,
                                                        snowflake_df=sf_input_dataframe
                                                        )
                except Exception as msg:
                    # return True, {"message": f"Failed to load model artifacts. {str(msg)}"}
                    raise Exception(f"Failed to load model artifacts. {str(msg)}")

                prob_columns =[]
                try:
                    prob_columns = list(
                    set(sf_input_dataframe.columns) - set(dataset_metadata['feature_names']) -set(dataset_metadata['target_column']) - set(dataset_metadata['prediction_column'])
                )
                except:
                    pass
                dataset_metadata['prob_columns'] = prob_columns
                    
                
                dataset_details['final_df'] = model_artifacts['final_df']
                dataset_details['hyper_parameters'] = model_artifacts['hyper_parameters']
                dataset_details['model_obj'] = model_artifacts['model_obj']

                temp_ui_data = {
                'dataset_type' : 'Dataset',
                'purpose' : 'Traning',
                'source' : dataset_source,
                'database': connection_params['database'],
                'schema': connection_params['schema'],
                'dataset_name': f'{dataset_name}_snowflake_training_dataset',
                'target_column': dataset_metadata['target_column'][0]
            }   
                dataset_ui_display.append(temp_ui_data)

                
            except Exception as msg:
                print(msg)
                return True, {"message": f"Failed to fetch dataset details from snowflake."}

        elif model_flavour.lower() == "sklearn" and str(type(model)).find("sklearn") > 0:
            try:
                dataset_metadata['feature_names'] = x_test.columns.to_list()
                dataset_metadata['target_column'] = [y_test.squeeze().name] if isinstance(y_train,pd.DataFrame) else [y_test.name]
                dataset_metadata['prediction_column'] = [y_pred.squeeze().name] if isinstance(y_pred,pd.DataFrame) else [y_train.name]
                if isinstance(prob,pd.DataFrame):
                    dataset_metadata['prob_columns'] = prob.columns.to_list()
                else:
                    dataset_metadata['prob_columns'] = []
                
                if x_train is not None:
                    x_train_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('x_train'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_x_train',
                    }
                    dataset_ui_display.append(x_train_data)

                if y_train is not None:
                    y_train_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('y_train'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_y_train',
                        'target_column': dataset_metadata['target_column'][0]
                    }
                    dataset_ui_display.append(y_train_data)

                if x_test is not None:
                    x_test_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('x_test'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_x_test'
                    }
                    dataset_ui_display.append(x_test_data)

                if y_test is not None:
                    y_test_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('y_test'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_y_test',
                        'target_column': dataset_metadata['target_column'][0]
                    }
                    dataset_ui_display.append(y_test_data)

                if y_pred is not None:
                    y_pred_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('y_pred'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_y_pred',
                        'pred_column': dataset_metadata['prediction_column'][0]
                    }
                    dataset_ui_display.append(y_pred_data)

                if prob is not None:
                    prob_data = {
                        'dataset_type' : 'Dataset',
                        'purpose' : purpose('prob'),
                        'source' : dataset_source,
                        'database': connection_params['database'],
                        'schema': connection_params['schema'],
                        'dataset_name': f'{dataset_name}_prob',
                        'prob_columns': dataset_metadata['prob_columns']
                    }
                    dataset_ui_display.append(prob_data)

                from fosforml.utils import load_model_artifacts
                try:
                    model_artifacts = load_model_artifacts(
                                                        session=session,
                                                        model_type=model_type,
                                                        model_obj=model,
                                                        snowflake_model=False,
                                                        x_test=x_test,
                                                        y_test=y_test,
                                                        y_pred=y_pred,
                                                        y_prob=prob
                                                        )
                except Exception as msg:
                    return True, {"message": f"Failed to load model artifacts. {str(msg)}"}
                    # raise Exception(f"Failed to load model artifacts. {str(msg)}")
                    
                dataset_details['final_df'] = model_artifacts['final_df']
                dataset_details['hyper_parameters'] = model_artifacts['hyper_parameters']
                dataset_details['model_obj'] = model_artifacts['model_obj']

            except Exception as msg:
                return True, {"message": f"Failed to get dataset details. {str(msg)}"}
                # raise Exception(f"Failed to get dataset details. {str(msg)}")  
        
        dataset_details['metadata'] = dataset_metadata
        dataset_details['ui_display'] = dataset_ui_display
        dataset_details['message'] = "Dataset details fetched successfully."
        
        return False,dataset_details


    def get_validation_df(self,df,no_rows):
        if df is not None and not df.empty:
            if isinstance(df,pd.DataFrame):
                return df.reset_index(drop=True).iloc[:no_rows,:]
            if isinstance(df,pd.Series):
                return df.to_frame().reset_index(drop=True).iloc[:no_rows,:]




from .snowflakesession import snowflakesession
from .model_registry import ModelRegistry
from .utilities import DatasetManager, Metadata
from .model_metrics import Classification, Regression
from .model import ModelObject

__all__ = [
    'snowflakesession',
    'ModelRegistry',
    'DatasetManager',
    'Metadata',
    'Classification',
    'Regression',
    'ModelObject'
]


import os,re
from .exceptions import ConfigError
from .constants import Headers, Client
from .constants import ModelConstants
import snowflake
import pandas as pd
from fosforml.model_manager.models import (SklearnModel,
                                           XGBoostModel,
                                           PipelineModel)

def get_model_name(name,source):
    """
    Get model name from the given name
    Experiment Models --- Experiment_<project id(please replace '-' with '_'>_FDC_<experiment_name>
    NOtebooks Models ---  Model_<project id(please replace '-' with '_'>_FDC_<model_name>
    """

    projectID = os.getenv("PROJECT_ID").upper().replace("-","_")

    if name is None:
        return False, "Model name is required"
    
    if not source or source.upper() == "NOTEBOOK":
        return True,f"MODEL_{projectID}_FDC_{name.upper()}"
    
    if source.upper() == "EXPERIMENT":
        from fosforml.constants import UserConfigs
        if UserConfigs.run_id in os.environ or UserConfigs.run_id.upper() in os.environ:
            run_id = os.getenv(UserConfigs.run_id, os.getenv(UserConfigs.run_id.upper()))
            return True, f"EXPERIMENT_{projectID}_FDC_{name.upper()}_{run_id}"
        else:
            print("RUN_ID is not found in environment variables")
            return False, f"EXPERIMENT_{projectID}_FDC_{name.upper()}"
    
    return False, "Invalid source"

def get_sf_dataset_columns_names(model_obj):
    """
    Get feature columns, target column and prediction columns for snowflake dataframes using snowflake model object
    """

    feature_names = model_obj.input_cols
    target_name = model_obj.label_cols
    prediction_column = model_obj.output_cols
    
    return feature_names, target_name, prediction_column

def read_jwt():
    config_file = os.path.expanduser(Client.config_file)
    if os.path.exists(config_file):
        with open(config_file) as f:
            jwt = f.read().strip("\n")
        return jwt
    elif os.getenv("TOKEN"):
        jwt = os.getenv("TOKEN")
        return jwt
    raise ConfigError


def get_headers():
    jwt = read_jwt()
    return {
        Headers.authorization: f"Token {jwt}",
        Headers.x_project_id: os.environ.get("PROJECT_ID"),
    }


def validate_dataset(dataset):
    if dataset is not None and not isinstance(dataset, (pd.DataFrame, pd.Series, snowflake.snowpark.dataframe.DataFrame)):
        raise ValueError("Dataset must be a DataFrame or Series")

def validate_column_names(df):
    pattern = r'^[A-Za-z_][A-Za-z_\d$]*$'
    if isinstance(df, pd.DataFrame):
        for col in df.columns.to_list():
            if not re.match(pattern, col):
                raise ValueError(f"Column name '{col}' is invalid. Column names must start with a letter or underscore and contain only letters, numbers, underscores and $ signs.")
        
    if isinstance(df,pd.Series):
        if not re.match(pattern, df.name):
            raise ValueError(f"Column name '{df.name}' is invalid. Column names must start with a letter or underscore and contain only letters, numbers, underscores and $ signs.")
        

def validate_datasets_for_sf(model_obj,name,model_type,flavour, snowflake_df, x_train, y_train, 
                             x_test, y_test, y_pred, prob, dataset_name, dataset_source,prediction_column):
        
    try:
        if not name:
            raise ValueError("Model name is required, please provide the value for 'name'")

        if not model_type:
            raise ValueError("Model type is required, please provide the value for 'model_type' as 'classification' or 'regression'")
        
        if not dataset_name:
            raise ValueError("Dataset name is required, please provide the value for 'dataset_name'")
        
        if not dataset_source:
            raise ValueError("Dataset source is required, please provide the value for 'dataset_source'")
    
        if flavour == ModelConstants.snowflake_model_flavour:
            if snowflake_df is None:
                raise ValueError("snowflake dataframe is required for flavour 'snowflake' with prediction columns,feature columns and target column")
            if not isinstance(snowflake_df, snowflake.snowpark.dataframe.DataFrame):
                raise ValueError("input dataframe must be a snowflake DataFrame")
            
            feature_names, target_name, prediction_col  = get_sf_dataset_columns_names(model_obj)
            
            if not set(feature_names).issubset(snowflake_df.columns):
                raise ValueError(f"Feature columns {feature_names} not found in snowflake dataframe")
            
            if not set(target_name).issubset(snowflake_df.columns):
                raise ValueError(f"Target column {target_name} not found in snowflake dataframe")
            
            if not prediction_col and prediction_column is None:
                raise ValueError("Prediction column is required for snowflake model, please provide the value for 'prediction_column'")
            
            if prediction_col and not set(prediction_col).issubset(snowflake_df.columns):
                raise ValueError(f"Prediction column {prediction_col} not found in snowflake dataframe")
            
            return True, "Snowflake Model datasets are validated successfully."

        else:
            if x_train is None or not isinstance(x_train, (pd.DataFrame)):
                raise ValueError("please provide 'x_train' as a pandas dataFrame")
            validate_column_names(x_train)

            if y_train is None or not isinstance(y_train, (pd.DataFrame, pd.Series)):
                raise ValueError("please provide 'y_train' as a pandas dataframe or series")
            validate_column_names(y_train)

            if x_test is None or not isinstance(x_test, (pd.DataFrame)):
                raise ValueError("please provide 'x_test' as a pandas dataFrame")
            validate_column_names(x_test)

            if y_test is None or not isinstance(y_test, (pd.DataFrame, pd.Series)):
                raise ValueError("please provide 'y_test' as a pandas dataframe or series")
            validate_column_names(y_test)

            if y_pred is None or not isinstance(y_pred, (pd.DataFrame, pd.Series)):
                raise ValueError("please provide 'y_pred' as a pandas dataframe or series")
            validate_column_names(y_pred)

            if prob is not None and not isinstance(prob, (pd.DataFrame)):
                raise ValueError("please provide 'prob' as a pandas dataframe class probabilities for each class ")
            validate_column_names(prob)
            
            return True, "Sklearn Model datasets are validated successfully."
        
    except Exception as e:
        return False,f"Failed to validate datasets: {str(e)}"


def load_model_artifacts(session,
                        model_type, 
                        model_obj,
                        snowflake_model=True,
                        snowflake_df=None,
                        x_test=None,
                        y_test=None,
                        x_train=None,
                        y_train=None,
                        y_pred=None,
                        y_prob=None,
                         ):
    """
    Load the model object from given snowflake model object
    """
    from fosforml.model_manager import ModelObject
    model_instance = ModelObject(model_obj,
                                model_type,
                                snowflake_model)
    return model_instance.get_model_artifacts(
            session=session,
            snowflake_df=snowflake_df,
            x_test=x_test,
            y_test=y_test,
            x_train=x_train,
            y_train=y_train,
            y_pred=y_pred,
            y_prob=y_prob
        )